import os

from ashdisperse import __file__ as _ad_file


def clean():
    _ad_dirc = os.path.dirname(_ad_file)

    for root, _, files in os.walk(_ad_dirc):
        for file in files:
            if file.endswith(".pyc") or file.endswith(".pyo") or file.endswith(".nbc") or file.endswith(".nbi"):
                file_path = os.path.join(root, file)
                try:
                    os.remove(file_path)
                except Exception as e:
                    print(f"Failed to delete {file_path}. Reason: {e}")


import os
from collections import OrderedDict
from datetime import datetime
from math import ceil, floor
from typing import Optional, Literal

import matplotlib.pyplot as plt
import metpy.calc as metcalc
import metpy.units as units
import numpy as np
import numpy.typing as npt
import pandas as pd
import xarray as xr
from matplotlib import ticker
from matplotlib.colors import Normalize
from netCDF4 import Dataset
from numba import float64, int64, jit, njit
from numba.experimental import jitclass
from numba.types import Tuple, unicode_type
from scipy.integrate import solve_ivp

from ashdisperse.met.met_netcdf import NetcdfMet
from ashdisperse.met.met_gfs import GFSarchive, GFSforecast
from ..params.params import Parameters_type
from ..queryreport import print_text
from ..utilities.utilities import (SA_Density_array, SA_Density_value,
                                   SA_Pressure_array, SA_Pressure_value,
                                   SA_Temperature_array, SA_Temperature_value,
                                   interp_ex_array, interp_ex_value)

MetData_spec = OrderedDict()
MetData_spec["z_data"] = float64[::1]
MetData_spec["temperature_data"] = float64[::1]
MetData_spec["pressure_data"] = float64[::1]
MetData_spec["density_data"] = float64[::1]
MetData_spec["wind_U_data"] = float64[::1]
MetData_spec["wind_V_data"] = float64[::1]
MetData_spec["source"] = unicode_type
MetData_spec["surface_temperature"] = float64  # surface temperature
MetData_spec["surface_pressure"] = float64  # surface pressure
MetData_spec["lapse_tropos"] = float64  # lapse rate in troposphere
MetData_spec["lapse_stratos"] = float64  # lapse rate stratosphere
MetData_spec["height_tropos"] = float64  # height of the troposphere
MetData_spec["height_stratos"] = float64  # height of the stratosphere
MetData_spec["Ra"] = float64  # gas constant of dry air
MetData_spec["g"] = float64  # gravitational acceleration


@jitclass(MetData_spec)
class MetData:
    def __init__(
        self,
        source: str ="standardAtmos",
        surface_temperature: float=293,
        surface_pressure: float=101325,
        lapse_tropos: float=6.5e-3,
        lapse_stratos: float=2.0e-3,
        height_tropos: float=11e3,
        height_stratos: float=20e3,
        Ra: float=285,
        g: float=9.81,
    ):
        self.surface_temperature = np.float64(surface_temperature)
        self.surface_pressure = np.float64(surface_pressure)
        self.lapse_tropos = np.float64(lapse_tropos)
        self.lapse_stratos = np.float64(lapse_stratos)
        self.height_tropos = np.float64(height_tropos)
        self.height_stratos = np.float64(height_stratos)
        self.Ra = np.float64(Ra)
        self.g = np.float64(g)
        self.source = source

        self.z_data = np.empty((0), dtype=np.float64)
        self.temperature_data = np.empty((0), dtype=np.float64)
        self.pressure_data = np.empty((0), dtype=np.float64)
        self.density_data = np.empty((0), dtype=np.float64)
        self.wind_U_data = np.empty((0), dtype=np.float64)
        self.wind_V_data = np.empty((0), dtype=np.float64)

    def set_zuv_data(self, z, u, v):
        self.z_data = z
        self.wind_U_data = u
        self.wind_V_data = v
            

    def temperature_array(self, z):
        if self.source in ["netCDF", "GFS", "GFS Archive"]:
            temperature = interp_ex_array(z, self.z_data, self.temperature_data)
        if self.source == "standardAtmos":
            temperature = SA_Temperature_array(
                z,
                self.surface_temperature,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
            )
        return temperature

    def temperature_value(self, z):
        if self.source in ["netCDF", "GFS", "GFS Archive"]:
            temperature = interp_ex_value(z, self.z_data, self.temperature_data)
        if self.source == "standardAtmos":
            temperature = SA_Temperature_value(
                z,
                self.surface_temperature,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
            )
        return temperature

    def pressure_array(self, z):
        if self.source in ["netCDF", "GFS", "GFS Archive"]:
            pressure = interp_ex_array(z, self.z_data, self.pressure_data)
        elif self.source == "standardAtmos":
            pressure = SA_Pressure_array(
                z,
                self.surface_temperature,
                self.surface_pressure,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
                self.g,
                self.Ra,
            )
        return pressure

    def pressure_value(self, z):
        if self.source in ["netCDF", "GFS", "GFS Archive"]:
            pressure = interp_ex_value(z, self.z_data, self.pressure_data)
        elif self.source == "standardAtmos":
            pressure = SA_Pressure_value(
                z,
                self.surface_temperature,
                self.surface_pressure,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
                self.g,
                self.Ra,
            )
        return pressure

    def density_array(self, z):
        if self.source in ["netCDF", "GFS", "GFS Archive"]:
            density = interp_ex_array(z, self.z_data, self.density_data)
        elif self.source == "standardAtmos":
            density = SA_Density_array(
                z,
                self.surface_temperature,
                self.surface_pressure,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
                self.g,
                self.Ra,
            )
        return density

    def density_value(self, z):
        if self.source in ["netCDF", "GFS", "GFS Archive"]:
            density = interp_ex_value(z, self.z_data, self.density_data)
        elif self.source == "standardAtmos":
            density = SA_Density_value(
                z,
                self.surface_temperature,
                self.surface_pressure,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
                self.g,
                self.Ra,
            )
        return density

    def wind_U_array(self, z, scale=None):
        u = interp_ex_array(z, self.z_data, self.wind_U_data)
        if scale is None:
            ret = u
        else:
            ret = u / scale
        return ret

    def wind_U_value(self, z, scale=None):
        u = interp_ex_value(z, self.z_data, self.wind_U_data)
        if scale is None:
            ret = u
        else:
            ret = u / scale
        return ret

    def wind_V_array(self, z, scale=None):
        v = interp_ex_array(z, self.z_data, self.wind_V_data)
        if scale is None:
            ret = v
        else:
            ret = v / scale
        return ret

    def wind_V_value(self, z, scale=None):
        v = interp_ex_value(z, self.z_data, self.wind_V_data)
        if scale is None:
            ret = v
        else:
            ret = v / scale
        return ret

    def wind_speed_array(self, z):
        u = self.wind_U_array(z)
        v = self.wind_V_array(z)
        spd = np.sqrt(u * u + v * v)
        return spd

    def wind_speed_value(self, z):
        u = self.wind_U_value(z)
        v = self.wind_V_value(z)
        spd = np.sqrt(u * u + v * v)
        return spd

    def max_wind_speed(self, H, num=100):
        z = np.arange(0.0, H + 1.0, H / (num - 1), dtype=np.float64)
        spd = self.wind_speed_array(z)
        return np.amax(spd)

    def settling_speed_array(self, params, z, scale=None):
        Ws = self.calculate_settling_speed_array(params, z)
        if scale is None:
            ret = Ws
        else:
            ret = Ws / scale
        return ret

    def settling_speed_value(self, params, z, scale=None):
        Ws = self.calculate_settling_speed_value(params, z)
        if scale is None:
            ret = Ws
        else:
            ret = Ws / scale
        return ret

    def settling_speed_for_grain_class_array(self, params, grain_i, z, scale=None):
        Ws = self.calculate_settling_speed_for_grain_class_array(params, grain_i, z)
        if scale is None:
            return Ws
        else:
            return Ws / scale

    def settling_speed_for_grain_class_value(self, params, grain_i, z, scale=None):
        Ws = self.calculate_settling_speed_for_grain_class_value(params, grain_i, z)
        if scale is None:
            return Ws
        else:
            return Ws / scale

    @staticmethod
    def _solve_settling_function(diam, g, rho_p, rho_a, mu, max_iter=50):

        tolerance = 1e-8

        x0 = np.float64(1e-6)
        x1 = np.float64(10000)

        def _settling_func_white(Re, d, g, rho_p, rho_a, mu):
            gp = (rho_p - rho_a) * g / rho_a
            C1 = 0.25
            C2 = 6.0
            Cd = C1 + 24.0 / Re + C2 / (1 + np.sqrt(Re))
            d3 = d**3
            f = Cd * Re * Re - 4.0 / 3.0 * gp * d3 * (rho_a / mu) ** 2
            return f

        fx0 = _settling_func_white(x0, diam, g, rho_p, rho_a, mu)
        fx1 = _settling_func_white(x1, diam, g, rho_p, rho_a, mu)

        if np.abs(fx0) < np.abs(fx1):
            x0, x1 = x1, x0
            fx0, fx1 = fx1, fx0

        x2, fx2 = x0, fx0
        d = x2

        mflag = True
        steps_taken = 0

        while steps_taken < max_iter and np.abs(x1 - x0) > tolerance:
            fx0 = _settling_func_white(x0, diam, g, rho_p, rho_a, mu)
            fx1 = _settling_func_white(x1, diam, g, rho_p, rho_a, mu)
            fx2 = _settling_func_white(x2, diam, g, rho_p, rho_a, mu)

            if fx0 != fx2 and fx1 != fx2:
                L0 = (x0 * fx1 * fx2) / ((fx0 - fx1) * (fx0 - fx2))
                L1 = (x1 * fx0 * fx2) / ((fx1 - fx0) * (fx1 - fx2))
                L2 = (x2 * fx1 * fx0) / ((fx2 - fx0) * (fx2 - fx1))
                new = L0 + L1 + L2
            else:
                new = x1 - fx1 * (x1 - x0) / (fx1 - fx0)

            if (
                (new < 0.25 * (3 * x0 + x1) or new > x1)
                or (mflag and np.abs(new - x1) >= 0.5 * np.abs(x1 - x2))
                or ((not mflag) and np.abs(new - x1) >= 0.5 * np.abs(x2 - d))
                or (mflag and np.abs(x1 - x2) < tolerance)
                or ((not mflag) and np.abs(x2 - d) < tolerance)
            ):
                new = 0.5 * (x0 + x1)
                mflag = True
            else:
                mflag = False

            fnew = _settling_func_white(new, diam, g, rho_p, rho_a, mu)
            d, x2 = x2, x1

            if fx0 * fnew < 0:
                x1 = new
            else:
                x0 = new

            if np.abs(fx0) < np.abs(fx1):
                x0, x1 = x1, x0

            steps_taken += 1

        return x1, steps_taken

    def calculate_settling_speed_array(self, params, z):

        ws = np.empty((z.size, params.grains.bins), dtype=np.float64)

        rho_a = np.empty((z.size, 1), dtype=np.float64)
        rho_a = self.density_array(z)

        for iz, rho_az in enumerate(rho_a):

            for j, (d, rho_p) in enumerate(
                zip(params.grains.diameter, params.grains.density)
            ):

                Re, steps_taken = self._solve_settling_function(
                    d, params.physical.g, rho_p, rho_az, params.physical.mu
                )
                if steps_taken > 50:
                    raise RuntimeError(
                        "In MetData: _solve_settling_function" + " failed to converge"
                    )
                ws[iz, j] = params.physical.mu * Re / rho_az / d

        return ws

    def calculate_settling_speed_value(self, params, z):

        ws = np.empty((params.grains.bins), dtype=np.float64)

        rho_a = self.density_value(z)

        for j, (d, rho_p) in enumerate(
            zip(params.grains.diameter, params.grains.density)
        ):

            Re, steps_taken = self._solve_settling_function(
                d, params.physical.g, rho_p, rho_a, params.physical.mu
            )
            if steps_taken > 50:
                raise RuntimeError(
                    "In MetData: _solve_settling_function" + " failed to converge"
                )
            ws[j] = params.physical.mu * Re / rho_a / d

        return ws

    def calculate_settling_speed_for_grain_class_array(self, params, grain_i, z):

        ws = np.empty((z.size, 1), dtype=np.float64)

        rho_a = np.empty((z.size, 1), dtype=np.float64)
        rho_a = self.density_array(z)

        for iz, rho_az in enumerate(rho_a):

            d = params.grains.diameter[grain_i]
            rho_p = params.grains.density[grain_i]

            Re, steps_taken = self._solve_settling_function(
                d, params.physical.g, rho_p, rho_az, params.physical.mu
            )
            if steps_taken > 50:
                raise RuntimeError(
                    "In MetData: _solve_settling_function" + " failed to converge"
                )
            ws[iz, 0] = params.physical.mu * Re / rho_az / d

        return ws

    def calculate_settling_speed_for_grain_class_value(self, params, grain_i, z):

        rho_a = self.density_value(z)

        d = params.grains.diameter[grain_i]
        rho_p = params.grains.density[grain_i]

        Re, steps_taken = self._solve_settling_function(
            d, params.physical.g, rho_p, rho_a, params.physical.mu
        )
        if steps_taken > 50:
            raise RuntimeError(
                "In MetData: _solve_settling_function" + " failed to converge"
            )
        ws = params.physical.mu * Re / rho_a / d

        return ws


# pylint: disable=E1101
MetData_type = MetData.class_type.instance_type


def save_met(met: MetData, file="meteorology.npz") -> None:
    if os.path.exists(file):
        print(
            "WARNING: {outname} ".format(outname=file)
            + "already exists and will be replaced"
        )

    if met.source == "standardAtmos":
        np.savez(
            file,
            source=met.source,
            surface_temperature=met.surface_temperature,
            surface_pressure=met.surface_pressure,
            lapse_tropos=met.lapse_tropos,
            lapse_stratos=met.lapse_stratos,
            height_tropos=met.height_tropos,
            height_stratos=met.height_stratos,
            Ra=met.Ra,
            g=met.g,
            z_data=met.z_data,
            wind_U_data=met.wind_U_data,
            wind_V_data=met.wind_V_data,
        )
    elif met.source in ["netCDF", "GFS"]:
        np.savez(
            file,
            source=met.source,
            Ra=met.Ra,
            g=met.g,
            z_data=met.z_data,
            wind_U_data=met.wind_U_data,
            wind_V_data=met.wind_V_data,
            temperature_data=met.temperature_data,
            pressure_data=met.pressure_data,
            density_data=met.density_data,
        )


def load_met(met_file: str) -> MetData:

    if not os.path.exists(met_file):
        raise IOError("AshDisperse meteorological file {} not found".format(met_file))

    met = MetData()
    data = np.load(met_file)

    if str(data["source"]) == "standardAtmos":
        met = MetData(
            source="standardAtmos",
            surface_temperature=np.float64(data["surface_temperature"]),
            surface_pressure=np.float64(data["surface_pressure"]),
            lapse_tropos=np.float64(data["lapse_tropos"]),
            lapse_stratos=np.float64(data["lapse_stratos"]),
            height_tropos=np.float64(data["height_tropos"]),
            height_stratos=np.float64(data["height_stratos"]),
            Ra=np.float64(data["Ra"]),
            g=np.float64(data["g"]),
        )
        met.set_zuv_data(data["Z"], data["U"], data["V"])
    elif str(data["source"]) in ["netCDF", "GFS"]:
        met = MetData(
            source="netCDF", Ra=np.float64(data["Ra"]), g=np.float64(data["g"])
        )
        met.z_data = data["z_data"]
        met.wind_U_data = data["wind_U_data"]
        met.wind_V_data = data["wind_V_data"]
        met.temperature_data = data["temperature_data"]
        met.pressure_data = data["pressure_data"]
        met.density_data = data["density_data"]

    return met

def load_met_from_repodf(met_dataset: pd.DataFrame, indx: int) -> MetData:

    if indx not in met_dataset.index:
        raise ValueError(f"index must be in met_dataset")
    
    met_data = met_dataset.iloc[indx]

    met = MetData()
    met.z_data = met_data['altitude']
    met.wind_U_data = met_data["wind_u"]
    met.wind_V_data = met_data["wind_v"]
    met.temperature_data = met_data["temperature"]
    met.pressure_data = met_data["pressure"]
    met.density_data = met_data["density"]

    return met


def netCDF_to_Met(met: MetData, netcdf_data: NetcdfMet) -> MetData:
    N = len(netcdf_data.altitude)
    met.z_data = np.empty((N), dtype=np.float64)
    met.z_data[:N] = netcdf_data.altitude[:N]
    met.wind_U_data = np.empty((N), dtype=np.float64)
    met.wind_U_data[:N] = netcdf_data.wind_U[:N]
    met.wind_V_data = np.empty((N), dtype=np.float64)
    met.wind_V_data[:N] = netcdf_data.wind_V[:N]
    met.temperature_data = np.empty((N), dtype=np.float64)
    met.temperature_data[:N] = netcdf_data.temperature[:N]
    met.pressure_data = np.empty((N), dtype=np.float64)
    met.pressure_data[:N] = netcdf_data.pressure[:N]
    met.density_data = np.empty((N), dtype=np.float64)
    met.density_data[:N] = netcdf_data.density[:N]
    met.source = "netCDF"
    return met


def gfs_to_Met(met: MetData, gfs_data: pd.DataFrame) -> MetData:
    gp = gfs_data["Geopotential_height_isobaric"]
    temp = gfs_data["Temperature_isobaric"]
    rh = gfs_data["Relative_humidity_isobaric"]
    u = gfs_data["ucomponent_of_wind_isobaric"]
    v = gfs_data["vcomponent_of_wind_isobaric"]
    pres = gfs_data["alt"]
    
    N = len(pres)

    # z = metcalc.

    met.z_data = np.empty((N), dtype=np.float64)
    met.z_data[:N] = np.flipud(z[:])
    # met.z_data[:N] = np.flipud(z[:].data.flatten())
    
    met.wind_U_data = np.empty((N), dtype=np.float64)
    met.wind_U_data[:N] = np.flipud(u[:])
    # met.wind_U_data[:N] = np.flipud(u[:].data.flatten())
    
    met.wind_V_data = np.empty((N), dtype=np.float64)
    met.wind_V_data[:N] = np.flipud(v[:])
    # met.wind_V_data[:N] = np.flipud(v[:].data.flatten())
    
    met.temperature_data = np.empty((N), dtype=np.float64)
    met.temperature_data[:N] = np.flipud(temp[:])
    # met.temperature_data[:N] = np.flipud(temp[:].data.flatten())
    
    met.pressure_data = np.empty((N), dtype=np.float64)
    met.pressure_data[:N] = np.flipud(pres[:])
    # met.pressure_data[:N] = np.flipud(pres[:].data.flatten())
    
    met.density_data = np.empty((N), dtype=np.float64)

    met.density_data[:N] = met.pressure_data[:N] / Ra / met.temperature_data[:N]
    # met.density_data[:N] = met.pressure_data[:N] / Ra / met.temperature_data[:N]
    
    met.source = "GFS"
    return met




def gfs_archive_to_Met(met: MetData, gfs_data: pd.DataFrame):
    z = gfs_data.altitude.values
    temp = gfs_data.temperature.values
    u = gfs_data.wind_U.values
    v = gfs_data.wind_V.values
    pres = gfs_data.pressure.values

    N = gfs_data.pressure.size

    met.z_data = np.empty((N), dtype=np.float64)
    met.z_data[:N] = z[:]

    met.wind_U_data = np.empty((N), dtype=np.float64)
    met.wind_U_data[:N] = u[:]

    met.wind_V_data = np.empty((N), dtype=np.float64)
    met.wind_V_data[:N] = v[:]

    met.temperature_data = np.empty((N), dtype=np.float64)
    met.temperature_data[:N] = temp[:]

    met.pressure_data = np.empty((N), dtype=np.float64)
    met.pressure_data[:N] = pres[:]

    met.density_data = np.empty((N), dtype=np.float64)
    met.density_data[:N] = met.pressure_data[:N] / Ra / met.temperature_data[:N]

    met.source = "GFS Archive"
    return met


def _near_lat_lon(target_lat: float, target_lon, lats, lons):

    lat_i_near = np.abs(lats - target_lat).argmin()
    lon_i_near = np.abs(lons - target_lon).argmin()

    if target_lat in lats:
        lat_i0 = lat_i_near
        lat_i1 = lat_i_near  # won't need this
    else:
        lat_i0 = lat_i_near if lat_i_near < target_lat else lat_i_near + 1
        lat_i1 = lat_i0 - 1
    if target_lon in lons:
        lon_i0 = lon_i_near
        lon_i1 = lon_i_near  # won't need this
    else:
        lon_i0 = lon_i_near if lon_i_near < target_lon else lon_i_near - 1
        lon_i1 = lon_i0 + 1

    return lat_i0, lat_i1, lon_i0, lon_i1


def _interp_latlon(target_lat, target_lon, data, lats, lons):

    lat_i0, lat_i1, lon_i0, lon_i1 = _near_lat_lon(target_lat, target_lon, lats, lons)

    d_lat = target_lat - lats[lat_i0]
    d_lon = target_lon - lons[lon_i0]

    data_00 = data[0, :, lat_i0, lon_i0]
    data_01 = data[0, :, lat_i1, lon_i0]
    data_10 = data[0, :, lat_i0, lon_i1]
    data_11 = data[0, :, lat_i1, lon_i1]

    data_i0 = data_00 + d_lon * (data_10 - data_00)
    data_i1 = data_01 + d_lon * (data_11 - data_01)
    data_ij = data_i0 + d_lat * (data_i1 - data_i0)

    return data_ij


def wind_scale(max_speed):

    max_log10 = np.log10(max_speed)
    major = []
    minor = []
    for j in range(np.int64(max_log10) + 1):
        major.append(10**j)
        for k in range(2, 10):
            if k * 10**j > max_speed:
                break
            minor.append(k * 10**j)

    return major, minor


def wind_plot(met, z, show=True, savename=None):

    fig, ax = plt.subplots()

    U = met.wind_U_array(z)
    V = met.wind_V_array(z)
    speed = met.wind_speed_array(z)

    max_speed = speed.max()
    major, minor = wind_scale(max_speed)
    major.append(minor[-1])
    minor.pop(-1)

    ax.set_xlim(-max_speed, max_speed)
    ax.set_ylim(-max_speed, max_speed)
    ax.set_aspect("equal")
    ax.axis("off")

    cmap = plt.cm.viridis
    norm = Normalize(vmin=0, vmax=z[-1])
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=Normalize(vmin=0, vmax=z[-1]))

    ax.quiver(
        0,
        -max_speed,
        0,
        2 * max_speed,
        color="darkgray",
        headwidth=6,
        headlength=10,
        scale=1,
        scale_units="xy",
        angles="xy",
    )
    ax.quiver(
        -max_speed / 2,
        0,
        max_speed,
        0,
        color="darkgray",
        headwidth=1,
        headlength=0,
        scale=1,
        scale_units="xy",
        angles="xy",
    )
    ax.text(0, 1.1 * max_speed, "N")

    for m in major:
        ax.add_artist(plt.Circle((0, 0), m, ec="darkgray", lw=0.5, fill=False))
        ax.text(m * np.cos(45 * np.pi / 180), m * np.sin(45 * np.pi / 180), str(m))
    for m in minor:
        ax.add_artist(plt.Circle((0, 0), m, ec="darkgray", lw=0.25, fill=False))

    for (this_z, this_u, this_v) in zip(z, U, V):
        ax.quiver(
            0,
            0,
            this_u,
            this_v,
            color=cmap(norm(this_z)),
            scale=1,
            scale_units="xy",
            angles="xy",
        )

    fmt = ticker.FuncFormatter(lambda z, pos: "{:g}".format(z * 1e-3))
    cbar = fig.colorbar(sm, format=fmt, ax=ax)
    cbar.ax.set_title("Altitude (km)")

    if savename is not None:
        plt.savefig(savename)

    if show:
        plt.show()
    else:
        plt.close()

    return




import os
import warnings
from math import ceil, floor

import branca.colormap as cm
import folium
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pyproj as Proj
import rasterio as rio
import rioxarray as rxa
import utm
import xarray as xa
from matplotlib import ticker
from matplotlib.colors import LogNorm, Normalize
from mpl_toolkits.axes_grid1 import make_axes_locatable
from pandas import concat
from rasterio.warp import Resampling, calculate_default_transform, reproject
from rasterio.windows import get_data_window
from scipy.fft import ifft2, irfft2
from scipy.interpolate import RectBivariateSpline, interp1d
from shapely.geometry import Polygon
from skimage.measure import find_contours, marching_cubes

from ashdisperse.params.emission_params import EmissionParameters
from ashdisperse.params.grain_params import GrainParameters
from ashdisperse.params.met_params import MetParameters
from ashdisperse.params.model_params import ModelParameters
from ashdisperse.params.output_params import OutputParameters
from ashdisperse.params.params import Parameters, copy_parameters
from ashdisperse.params.physical_params import PhysicalParameters
from ashdisperse.params.solver_params import SolverParameters
from ashdisperse.params.source_params import SourceParameters
from ashdisperse.spectral import grid_freq
from ashdisperse.version import __version__

from ..mapping import (BindColormap, add_north_arrow, add_opentopo_basemap,
                       add_scale_bar, ax_ticks, latlon_to_utm_epsg,
                       set_figure_size, set_min_axes, stamen_zoom_level,
                       webmerc, wgs84)
from ..utilities import (lin_levels, log_levels, log_steps, nice_round_down,
                         nice_round_up, pad_window, plot_rowscols)


def compat_warning(message, category, filename, lineno, file=None, line=None):
    return f"{category.__name__}: {message}"

def _clip_to_window(raster, x, y, vmin=1e-5, pad=1):
    height, width = raster.shape
    raster_tmp = np.copy(raster)
    raster_tmp[raster < vmin] = -1.0
    window = get_data_window(raster_tmp, nodata=-1.0)
    winpad = pad_window(window, height, width, pad=pad)

    row_start = winpad["row_start"]
    row_stop = winpad["row_stop"]
    col_start = winpad["col_start"]
    col_stop = winpad["col_stop"]
    raster_out = raster[row_start:row_stop, col_start:col_stop]
    x_out = x[col_start:col_stop]
    y_out = y[row_start:row_stop]
    return raster_out, x_out, y_out


def _interpolate(raster, x, y, resolution, extent=None, nodata=-1, vmin=1e-6):

    if extent is not None:
        min_x = extent[0]
        min_y = extent[1]
        max_x = extent[2]
        max_y = extent[3]
    else:
        raster_tmp = np.copy(raster)
        raster_tmp[raster < vmin] = nodata
        window = get_data_window(raster_tmp, nodata=nodata)
        min_x = x[window.col_off]
        min_y = y[window.row_off]
        max_x = x[window.col_off + window.width - 1]
        max_y = y[window.row_off + window.height - 1]

    print("resolution = ", resolution)

    left = floor(min_x / resolution) * resolution
    right = ceil(max_x / resolution) * resolution
    bottom = floor(min_y / resolution) * resolution
    top = ceil(max_y / resolution) * resolution

    width = int((right - left) / resolution) + 1
    height = int((top - bottom) / resolution) + 1
    print("Output raster shape is ({h},{w})".format(h=height, w=width))

    x_out = np.linspace(left, right, num=width, endpoint=True)
    y_out = np.linspace(bottom, top, num=height, endpoint=True)

    raster_tmp, x_tmp, y_tmp = _clip_to_window(raster, x, y, vmin=vmin / 10, pad=12)
    raster_tmp[raster_tmp < vmin / 10] = 0.0

    intrp = RectBivariateSpline(y_tmp, x_tmp, raster_tmp)

    raster_out = intrp(y_out, x_out)

    return raster_out, x_out, y_out


class AshDisperseResult:
    """Results of an AshDisperse simulation"""
    def __init__(self, params, C0_FT, Cz_FT):
        """Create AshDisperse result class"""
        self.grain_classes = params.grains.bins
        self.params = copy_parameters(params)

        self._interpolated = False
        self._changed_MER = False
        self._changed_gsd = False
        self._changed_duration = False

        self.utm = utm.from_latlon(
            self.params.source.latitude, self.params.source.longitude
        )

        self.utmepsg = latlon_to_utm_epsg(
            self.params.source.latitude, self.params.source.longitude
        )

        self.C0_FT = C0_FT
        self.Cz_FT = Cz_FT

        Nx = self.params.solver.Nx
        Ny = self.params.solver.Ny
        Ng = self.params.grains.bins
        Nz = self.params.output.Nz

        x, kx = grid_freq(Nx)
        y, ky = grid_freq(Ny)

        self.x_dimless = x
        self.y_dimless = y

        self.kx = kx
        self.ky = ky

        res_x = x[1] - x[0]
        res_y = y[1] - y[0]

        C0 = np.zeros((Ny, Nx, Ng), dtype=np.float64)
        Cz = np.zeros((Ny, Nx, Nz, Ng), dtype=np.float64)
        
        # for igrain in range(Ng):
        #     C0[:, :, igrain] = np.real(ifft2(C0_FT[:, :, igrain], s=(Ny,Nx))) / (res_x * res_y)
        #     for k in range(Nz):
        #         Cz[:, :, k, igrain] = np.real(ifft2(Cz_FT[:, :, k, igrain], s=(Ny,Nx))) / (res_x * res_y)
        for igrain in range(Ng):
            C0[:, :, igrain] = irfft2(C0_FT[:, :, igrain], s=(Ny,Nx)) / (res_x * res_y)
            for k in range(Nz):
                Cz[:, :, k, igrain] = irfft2(Cz_FT[:, :, k, igrain], s=(Ny,Nx)) / (res_x * res_y)

        self.C0_dimless = C0
        self.Cz_dimless = Cz

        self.C0 = np.zeros_like(C0)
        self.Cz = np.zeros_like(Cz)
        self.SettlingFlux = np.zeros_like(C0)

        self.x = np.zeros((len(x), params.grains.bins))
        self.y = np.zeros((len(y), params.grains.bins))

        for j in range(params.grains.bins):
            # self.x[:, j] = x * params.model.xyScale[j] * params.model.Lx[j] / np.pi
            # self.y[:, j] = y * params.model.xyScale[j] * params.model.Ly[j] / np.pi
            self.x[:, j] = x * params.model.xScale[j] * params.model.Lx[j] / np.pi
            self.y[:, j] = y * params.model.yScale[j] * params.model.Ly[j] / np.pi
            self.C0[:, :, j] = C0[:, :, j] * params.model.cScale[j]
            self.Cz[:, :, :, j] = Cz[:, :, :, j] * params.model.cScale[j]
            self.SettlingFlux[:, :, j] = (
                params.met.Ws_scale[j] * C0[:, :, j] * params.model.cScale[j]
            )

        df = pd.DataFrame(columns=["Name", "Latitude", "Longitude"])
        df.loc[0] = [
            self.params.source.name,
            self.params.source.latitude,
            self.params.source.longitude,
        ]
        self.source_marker = gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(df.Longitude, df.Latitude),
            crs=wgs84["init"],
        )

    def linear_interp(self, Nx_log2: int, Ny_log2: int):
        "Do a bilinear interpolation by zero-padding the FFTs"
        
        Nx_old = self.params.solver.Nx
        Ny_old = self.params.solver.Ny
        Ng = self.params.grains.bins
        Nz = self.params.output.Nz

        Nx = 2**Nx_log2
        Ny = 2**Ny_log2

        half_Nx_old = Nx_old//2 + 1
        half_Nx = Nx//2 + 1

        half_Ny_old = Ny_old//2

        C0_FT = np.zeros((Ny, half_Nx, Ng), dtype=np.complex128)
        Cz_FT = np.zeros((Ny, half_Nx, Nz, Ng), dtype=np.complex128)

        C0_FT[:half_Ny_old,:half_Nx_old,:] = self.C0_FT[:half_Ny_old,:,:]
        C0_FT[Ny-half_Ny_old+1:,:half_Nx_old, :] = self.C0_FT[half_Ny_old+1:,:, :]

        Cz_FT[:half_Ny_old,:half_Nx_old,:, :] = self.Cz_FT[:half_Ny_old,:,:, :]
        Cz_FT[Ny-half_Ny_old+1:,:half_Nx_old, :, :] = self.Cz_FT[half_Ny_old+1:,:, :, :]

        params = copy_parameters(self.params)
        params.solver.Nx_log2 = Nx_log2
        params.solver.Ny_log2 = Ny_log2

        r = AshDisperseResult(params, C0_FT, Cz_FT)
        r._interpolated = True

        return r


    def change_MER(self,MER):
        params = self.params
        params.source.MER = MER
        self.params.model.from_params(
            params,
            params.model.xScale,
            params.model.yScale
        )

        self.__init__(params, self.C0_FT, self.Cz_FT)
        self._changed_MER = True

    def change_duration(self,duration):
        self.params.source.duration = duration
        self._changed_duration = True

    def change_grain_proportions(self,props):
        if len(props)!=self.params.grains.bins:
            raise RuntimeError(f'length of props must equal number of grain size classes ({self.params.grains.bins})')

        params = self.params
        for j, p in enumerate(props):
            params.grains.proportion[j] = np.float64(p)

        self.params.model.from_params(
            params,
            params.model.xScale,
            params.model.yScale
        )

        self.__init__(params, self.C0_FT, self.Cz_FT)
        self._changed_gsd = True

    @classmethod
    def from_netcdf(cls, filename):
        """Initialize results from a netcdf file"""
        da = xa.load_dataset(filename)

        if da.version != __version__:
            warnings.formatwarning = compat_warning
            warnings.warn(f"Loading AshDisperse results from version {da.version}. Current version is {__version__}. \n There may be incompatibilities.\n")

        p = Parameters()
        p.solver = SolverParameters(
            domX=da.domX,
            domY=da.domY,
            minN_log2=da.minN_log2,
            maxN_log2=da.maxN_log2,
            Nx_log2=da.Nx_log2,
            Ny_log2=da.Ny_log2,
            epsilon=da.epsilon,
        )
        p.grains = GrainParameters()
        # TODO: do this properly
        if isinstance(da.diameter, np.float64):
            p.grains.from_values(da.diameter, da.density, da.proportion)
        else:
            p.grains.from_lists(da.diameter, da.density, da.proportion)

        p.source = SourceParameters(
            latitude=da.latitude,
            longitude=da.longitude,
            utmcode=da.utmcode,
            radius=da.radius,
            PlumeHeight=da.PlumeHeight,
            MER=da.MER,
            duration=da.duration,
            name=da.name,
        )

        p.emission = EmissionParameters()
        # TODO: do this properly
        if isinstance(da.emission_lower, np.float64):
            p.emission.from_values(da.emission_lower,
                da.emission_upper,
                da.emission_profile,
                da.Suzuki_k)
        else:
            p.emission.from_lists(
                da.emission_lower,
                da.emission_upper, 
                da.emission_profile, 
                da.Suzuki_k)

        p.physical = PhysicalParameters(
            Kappa_h=da.Kappa_h,
            Kappa_v=da.Kappa_v,
            g=da.g,
            mu=da.mu,
        )

        p.met = MetParameters(da.U_scale, np.atleast_1d(da.Ws_scale))

        p.output = OutputParameters(start=da.start, stop=da.stop, step=da.step)
        p.output.ChebMats(p.solver.maxN, p.source.PlumeHeight)

        p.model = ModelParameters()
        if isinstance(da.SettlingScale, np.float64):
            p.model.from_values(
                da.SettlingScale,
                da.Velocity_ratio,
                da.xScale,
                da.yScale,
                da.Lx,
                da.Ly,
                da.cScale,
                da.QScale,
                da.Peclet_number,
                da.Diffusion_ratio,
                da.sigma_hat,
                da.sigma_hat_scale,
            )
        else:
            p.model.from_lists(
                da.SettlingScale,
                da.Velocity_ratio,
                da.xScale,
                da.yScale,
                da.Lx,
                da.Ly,
                da.cScale,
                da.QScale,
                da.Peclet_number,
                da.Diffusion_ratio,
                da.sigma_hat,
                da.sigma_hat_scale,
            )

        C0_FT = da.C0_FT_r.values + 1j*da.C0_FT_i.values
        Cz_FT = da.Cz_FT_r.values + 1j*da.Cz_FT_i.values

        cls = AshDisperseResult(p, C0_FT, Cz_FT)

        return cls

    def to_netcdf(self, filename="AshDisperse.nc", compress=True, complevel=5):
        """Save AshDisperse results to a netCDF file"""
        Nx = self.params.solver.Nx
        da = xa.Dataset(
            data_vars=dict(
                C0_FT_r=(["ky", "kx", "grains"], np.real(self.C0_FT)),
                C0_FT_i=(["ky", "kx", "grains"], np.imag(self.C0_FT)),
                Cz_FT_r=(["ky", "kx", "z", "grains"], np.real(self.Cz_FT)),
                Cz_FT_i=(["ky", "kx", "z", "grains"], np.imag(self.Cz_FT)),
            ),
            coords=dict(
                grains=np.arange(self.params.grains.bins),
                kx=self.kx[:Nx//2+1],
                ky=self.ky,
                z=self.params.output.altitudes,
            ),
            attrs=dict(
                description="AshDisperse results",
                version=__version__,
                domX=self.params.solver.domX,
                domY=self.params.solver.domY,
                minN_log2=self.params.solver.minN_log2,
                maxN_log2=self.params.solver.maxN_log2,
                Nx_log2=self.params.solver.Nx_log2,
                Ny_log2=self.params.solver.Ny_log2,
                epsilon=self.params.solver.epsilon,
                meps=self.params.solver.meps,
                bins=self.params.grains.bins,
                diameter=list(self.params.grains.diameter),
                density=list(self.params.grains.density),
                proportion=list(self.params.grains.proportion),
                name=self.params.source.name,
                latitude=self.params.source.latitude,
                longitude=self.params.source.longitude,
                utmcode=self.params.source.utmcode,
                radius=self.params.source.radius,
                PlumeHeight=self.params.source.PlumeHeight,
                MER=self.params.source.MER,
                duration=self.params.source.duration,
                emission_lower=list(self.params.emission.lower),
                emission_upper=list(self.params.emission.upper),
                emission_profile=list(self.params.emission.profile),
                Suzuki_k=list(self.params.emission.Suzuki_k),
                Kappa_h=self.params.physical.Kappa_h,
                Kappa_v=self.params.physical.Kappa_v,
                g=self.params.physical.g,
                mu=self.params.physical.mu,
                U_scale=self.params.met.U_scale,
                Ws_scale=list(self.params.met.Ws_scale),
                SettlingScale=list(self.params.model.SettlingScale),
                Velocity_ratio=list(self.params.model.Velocity_ratio),
                xScale=list(self.params.model.xScale),
                yScale=list(self.params.model.yScale),
                Lx=list(self.params.model.Lx),
                Ly=list(self.params.model.Ly),
                cScale=list(self.params.model.cScale),
                QScale=list(self.params.model.QScale),
                Peclet_number=self.params.model.Peclet_number,
                Diffusion_ratio=self.params.model.Diffusion_ratio,
                sigma_hat=list(self.params.model.sigma_hat),
                sigma_hat_scale=list(self.params.model.sigma_hat_scale),
                start=self.params.output.start,
                stop=self.params.output.stop,
                step=self.params.output.step,
                altitudes=self.params.output.altitudes,
                Nz=self.params.output.Nz,
            ),
        )

        if compress:
            comp = dict(zlib=True, complevel=complevel)
            #comp = {"compression": "gzip", "compression_opts": complevel}
            encoding = {var: comp for var in da.data_vars}
        else:
            encoding = {}

        da.to_netcdf(
            path=filename,
            mode="w",
            format="NETCDF4",
            engine="h5netcdf",
            encoding=encoding,
        )

        return

    def _check_valid_grain(self, grain_i):
        if grain_i > self.params.grains.bins:
            raise ValueError(
                "Requested grain class not found; given {}".format(grain_i)
                + " in results with"
                + " {} grain classes.".format(self.params.grains.bins)
            )
        if grain_i < 0:
            raise ValueError("Grain class index must be positive")


    def get_groundconc_for_grain_class(self, grain_i, vmin=1e-9, masked=False):
        self._check_valid_grain(grain_i)

        x = self.x[:, grain_i]
        y = self.y[:, grain_i]

        conc = self.C0[:, :, grain_i]

        conc, x, y = _clip_to_window(conc, x, y, vmin=vmin)

        if conc.size == 0:
            return None

        if np.nanmax(conc) < vmin:
            return None

        if masked:
            conc = np.ma.masked_less(conc, vmin)

        return x, y, conc


    def get_settlingflux_for_grain_class(self, grain_i, vmin=1e-7, masked=False):
        self._check_valid_grain(grain_i)

        x = self.x[:, grain_i]
        y = self.y[:, grain_i]

        flux = self.SettlingFlux[:, :, grain_i]

        flux, x, y = _clip_to_window(flux, x, y, vmin=vmin)

        if flux.size == 0:
            return None

        if np.nanmax(flux) < vmin:
            return None

        if masked:
            flux = np.ma.masked_less(flux, vmin)

        return x, y, flux

    def get_ashload_for_grain_class(self, grain_i, vmin=1e-5, masked=False, clipped=True):
        self._check_valid_grain(grain_i)

        x = self.x[:, grain_i]
        y = self.y[:, grain_i]

        ashload = self.SettlingFlux[:, :, grain_i] * self.params.source.duration

        if clipped:
            ashload, x, y = _clip_to_window(ashload, x, y, vmin=vmin)

        if ashload.size == 0:
            return None

        if np.nanmax(ashload) < vmin:
            return None

        if masked:
            ashload = np.ma.masked_less(ashload, vmin)

        return x, y, ashload

    def POI_ashload_for_grain_class(
        self, grain_i: int, lat: float, lon: float, vmin: float = 1e-3
    ):
        self._check_valid_grain(grain_i)

        df = pd.DataFrame(columns=["Name", "Latitude", "Longitude"])
        df.loc[0] = ["POI", lat, lon]
        POI = gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(df.Longitude, df.Latitude),
            crs=wgs84["init"],
        )

        POI = POI.to_crs(self.utmepsg)

        try:
            ashload = self.raster_ashload_for_grain_class(
                grain_i, vmin=vmin, nodata=np.nan, masked=False, crs=None
            )

            POI_ashload = ashload.interp(
                x=POI.geometry.x, y=POI.geometry.y, method="cubic"
            )

            load = POI_ashload["ash_load"].values[0][0]

            if np.isnan(load):
                return 0
            elif load < 0:
                return 0
            else:
                return load
        except:
            return 0

    def POI_ashload(self, lat: float, lon: float, vmin: float = 1e-3):

        loads = []
        total_load = 0
        for j in range(self.grain_classes):
            this_ashload = self.POI_ashload_for_grain_class(j, lat, lon, vmin)
            loads.append(
                {
                    "diameter": self.params.grains.diameter[j],
                    "density": self.params.grains.density[j],
                    "proportion": self.params.grains.proportion[j],
                    "load": this_ashload,
                }
            )
            total_load += this_ashload

        return total_load, loads

    def POI_groundconc_for_grain_class(
        self, grain_i: int, lat: float, lon: float, vmin: float = 1e-3
    ):
        self._check_valid_grain(grain_i)

        df = pd.DataFrame(columns=["Name", "Latitude", "Longitude"])
        df.loc[0] = ["POI", lat, lon]
        POI = gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(df.Longitude, df.Latitude),
            crs=wgs84["init"],
        )

        POI = POI.to_crs(self.utmepsg)

        try:
            
            conc = self.raster_groundconc_for_grain_class(

                grain_i, vmin=vmin, nodata=np.nan, masked=False, crs=None
            )

            POI_ashload = conc.interp(
                x=POI.geometry.x, y=POI.geometry.y, method="cubic"
            )

            load = POI_ashload["ground_concentration"].values[0][0]

            if np.isnan(load):
                return 0
            elif load < 0:
                return 0
            else:
                return load
        except:
            return 0

    def POI_groundconc(self, lat: float, lon: float, vmin: float = 1e-3):

        concs = []
        total_conc = 0
        for j in range(self.grain_classes):
            this_conc = self.POI_groundconc_for_grain_class(j, lat, lon, vmin)
            concs.append(
                {
                    "diameter": self.params.grains.diameter[j],
                    "density": self.params.grains.density[j],
                    "proportion": self.params.grains.proportion[j],
                    "ground_concentration": this_conc,
                }
            )
            total_conc += this_conc

        return total_conc, concs
    
    def POI_settlingflux_for_grain_class(
        self, grain_i: int, lat: float, lon: float, vmin: float = 1e-3
    ):
        self._check_valid_grain(grain_i)

        df = pd.DataFrame(columns=["Name", "Latitude", "Longitude"])
        df.loc[0] = ["POI", lat, lon]
        POI = gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(df.Longitude, df.Latitude),
            crs=wgs84["init"],
        )

        POI = POI.to_crs(self.utmepsg)

        try:
            
            flux = self.raster_settlingflux_for_grain_class(

                grain_i, vmin=vmin, nodata=np.nan, masked=False, crs=None
            )

            POI_flux = flux.interp(
                x=POI.geometry.x, y=POI.geometry.y, method="cubic"
            )

            flux = POI_flux["settling_flux"].values[0][0]

            if np.isnan(flux):
                return 0
            elif flux < 0:
                return 0
            else:
                return flux
        except:
            return 0

    def POI_settlingflux(self, lat: float, lon: float, vmin: float = 1e-3):

        fluxes = []
        total_flux = 0
        for j in range(self.grain_classes):
            this_flux = self.POI_settlingflux_for_grain_class(j, lat, lon, vmin)
            fluxes.append(
                {
                    "diameter": self.params.grains.diameter[j],
                    "density": self.params.grains.density[j],
                    "proportion": self.params.grains.proportion[j],
                    "settling_flux": this_flux,
                }
            )
            total_flux += this_flux

        return total_flux, fluxes

    def max_ashload_for_grain_class(self, grain_i):
        self._check_valid_grain(grain_i)
        return np.nanmax(self.SettlingFlux[:, :, grain_i]) * self.params.source.duration


    def raster_groundconc_for_grain_class(
        self, grain_i, vmin=1e-9, nodata=np.nan, masked=False, crs=None
    ):

        x, y, conc = self.get_groundconc_for_grain_class(
            grain_i, vmin=vmin, masked=masked
        )

        ds = xa.Dataset()
        ds.coords["x"] = x + self.utm[0]
        ds.coords["y"] = y + self.utm[1]

        ds["ground_concentration"] = (("y", "x"), conc)
        ds["ground_concentration"].attrs["short_name"] = "ground_concentration_{}".format(grain_i)
        ds["ground_concentration"].attrs[
            "long_name"
        ] = "concentration at ground level for grain size {} m".format(
            self.params.grains.diameter[grain_i]
        )
        ds["ground_concentration"].attrs["units"] = "kg/m**3"
        ds["ground_concentration"].rio.write_nodata(nodata, inplace=True)

        ds.x.attrs["units"] = "metres"
        ds.y.attrs["units"] = "metres"

        ds = ds.rio.write_crs(self.utmepsg)

        if crs is not None:
            ds = ds.rio.reproject(crs, resampling=Resampling.cubic)

        return ds


    def raster_settlingflux_for_grain_class(
        self, grain_i, vmin=1e-6, nodata=np.nan, masked=False, crs=None
    ):

        x, y, flux = self.get_settlingflux_for_grain_class(
            grain_i, vmin=vmin, masked=masked
        )

        ds = xa.Dataset()
        ds.coords["x"] = x + self.utm[0]
        ds.coords["y"] = y + self.utm[1]

        ds["settling_flux"] = (("y", "x"), flux)
        ds["settling_flux"].attrs["short_name"] = "settling_flux_{}".format(grain_i)
        ds["settling_flux"].attrs[
            "long_name"
        ] = "settling flux at ground level for grain size {} m".format(
            self.params.grains.diameter[grain_i]
        )
        ds["settling_flux"].attrs["units"] = "kg/m**2/s"
        ds["settling_flux"].rio.write_nodata(nodata, inplace=True)

        ds.x.attrs["units"] = "metres"
        ds.y.attrs["units"] = "metres"

        ds = ds.rio.write_crs(self.utmepsg)

        if crs is not None:
            ds = ds.rio.reproject(crs, resampling=Resampling.cubic)

        return ds

    def raster_ashload_for_grain_class(
        self, grain_i, vmin=1e-2, nodata=np.nan, masked=False, clipped=True, crs=None
    ):

        x, y, ashload = self.get_ashload_for_grain_class(
            grain_i, vmin=vmin, masked=masked, clipped=clipped,
        )

        ds = xa.Dataset()
        ds.coords["x"] = x + self.utm[0]
        ds.coords["y"] = y + self.utm[1]

        ds["ash_load"] = (("y", "x"), ashload)
        ds["ash_load"].attrs["short_name"] = "ash_load_{}".format(grain_i)
        ds["ash_load"].attrs[
            "long_name"
        ] = "ash load at ground level for grain size {} m".format(
            self.params.grains.diameter[grain_i]
        )
        ds["ash_load"].attrs["units"] = "kg/m**2"
        ds["ash_load"].rio.write_nodata(nodata, inplace=True)

        ds.x.attrs["units"] = "metres"
        ds.y.attrs["units"] = "metres"

        ds = ds.rio.write_crs(self.utmepsg)

        if crs is not None:
            ds = ds.rio.reproject(crs, resampling=Resampling.cubic, nodata=nodata)

        return ds
    

    def _raster_contour(self, raster, name, cntrs):

        data = raster.data
        x = raster.x.data
        y = raster.y.data
        crs = raster.rio.crs

        data_min = np.nanmin(data)
        data_max = np.nanmax(data)
        data_max = nice_round_up(data_max, mag=10 ** np.floor(np.log10(data_max)))

        fx = interp1d(np.arange(0, len(x)), x)
        fy = interp1d(np.arange(0, len(y)), y)

        cntrs = cntrs[cntrs > data_min]
        cntrs = cntrs[cntrs < data_max]

        for kk, this_cntr in enumerate(cntrs):

            C = find_contours(data, this_cntr)

            if len(C) == 0:
                pass
            for jj, p in enumerate(C):
                p[:, 0] = fy(p[:, 0])
                p[:, 1] = fx(p[:, 1])

                p[:, [0, 1]] = p[:, [1, 0]]

                if len(p[:, 1]) > 2:
                    thisPoly = Polygon(p).buffer(0)
                    if not thisPoly.is_empty:
                        if jj == 0:
                            geom = thisPoly
                            g_tmp = gpd.GeoDataFrame(
                                columns=["contour", "name", "geometry"], crs=crs
                            )
                            g_tmp.loc[0, "contour"] = this_cntr
                            g_tmp.loc[[0], "geometry"] = gpd.GeoSeries([geom])
                            g_tmp.loc[0, "name"] = name
                        else:
                            if g_tmp.loc[0, "geometry"].contains(thisPoly):
                                geom = g_tmp.loc[0, "geometry"].difference(thisPoly)
                            else:
                                try:
                                    geom = g_tmp.loc[0, "geometry"].union(thisPoly)
                                except:
                                    print(
                                        "Error processing polygon contour -- skipping.  Better check the result!"
                                    )
                            g_tmp.loc[[0], "geometry"] = gpd.GeoSeries([geom])

            if kk == 0:
                g = gpd.GeoDataFrame(g_tmp).set_geometry("geometry")
            else:
                g = gpd.GeoDataFrame(pd.concat([g, g_tmp], ignore_index=True))

        g["contour"] = g["contour"].astype("float64")
        g.crs = crs

        return g

    def write_gtiff(self, raster, x, y, outname, nodata=-1, vmin=1e-6, resolution=None):

        print("Writing data to {outname}".format(outname=outname))
        if os.path.isfile(outname):
            print(
                "WARNING: {outname} ".format(outname=outname)
                + "already exists and will be replaced"
            )

        utmcode = self.params.source.utmcode

        if resolution is not None:
            raster, x, y = _interpolate(
                raster, x, y, resolution, nodata=nodata, vmin=vmin
            )

        height, width = raster.shape
        raster[raster < vmin] = nodata

        transform = rio.transform.from_bounds(
            x[0] + self.utm[0],
            y[0] + self.utm[1],
            x[-1] + self.utm[0],
            y[-1] + self.utm[1],
            width,
            height,
        )

        with rio.open(
            outname,
            "w",
            driver="GTiff",
            height=height,
            width=width,
            count=1,
            dtype=str(raster.dtype),
            crs=utmcode,
            nodata=nodata,
            transform=transform,
            resampling=Resampling.cubic,
            compress="lzw",
        ) as dst:
            dst.write(np.flipud(raster), 1)

    def write_settling_flux_for_grain_class(
        self, grain_i, nodata=-1, vmin=1e-6, resolution=None
    ):
        self._check_valid_grain(grain_i)
        self.write_gtiff(
            self.SettlingFlux[:, :, grain_i],
            self.x[:, grain_i],
            self.y[:, grain_i],
            "SettlingFlux_{}.tif".format(grain_i),
            nodata=nodata,
            vmin=vmin,
            resolution=resolution,
        )

    def write_ashload_for_grain_class(
        self, grain_i, vmin=1e-3, resolution=None, outname=None, compress='LZW',
    ):
        self._check_valid_grain(grain_i)

        ashload = self.raster_ashload_for_grain_class(grain_i, vmin=vmin, masked=False)

        res_x, res_y = ashload.rio.resolution()
        height = ashload.rio.height
        width = ashload.rio.width
        if resolution is not None:
            new_width = round(width * res_x / resolution)
            new_height = round(height * res_y / resolution)
            ashload = ashload.rio.reproject(
                ashload.rio.crs,
                shape=(new_height, new_width),
                resampling=Resampling.cubic,
            )

        if outname is None:
            outname = "AshLoad_{}.tif".format(grain_i)

        print("Writing data to {outname}".format(outname=outname))
        if os.path.isfile(outname):
            print(
                "WARNING: {outname} ".format(outname=outname)
                + "already exists and will be replaced"
            )

        ashload.rio.to_raster(outname, compress=compress)
        return

    def get_ashload(
        self,
        resolution=300.0,
        vmin=1e-3,
        nodata=-1,
        export_gtiff=True,
        export_name="AshLoad.tif",
    ):

        ash_load_grains = self.SettlingFlux * self.params.source.duration
        Ngrains = self.params.grains.bins

        row_off = []
        col_off = []
        h = []
        w = []
        x_min = []
        y_min = []
        x_max = []
        y_max = []

        for igrain in range(Ngrains):
            thisAshLoad = ash_load_grains[:, :, igrain]
            thisAshLoad[thisAshLoad < vmin / Ngrains / 10] = nodata
            window = get_data_window(thisAshLoad, nodata=nodata)
            row_off.append(window.row_off)
            col_off.append(window.col_off)
            h.append(window.height)
            w.append(window.width)
            x_min.append(self.x[window.col_off, igrain])
            x_max.append(self.x[window.col_off + window.width - 1, igrain])
            y_min.append(self.y[window.row_off, igrain])
            y_max.append(self.y[window.row_off + window.height - 1, igrain])

        min_x = min(x_min)
        min_y = min(y_min)
        max_x = max(x_max)
        max_y = max(y_max)

        left = floor(min_x / resolution) * resolution
        right = ceil(max_x / resolution) * resolution
        bottom = floor(min_y / resolution) * resolution
        top = ceil(max_y / resolution) * resolution

        width = int((right - left) / resolution) + 1
        height = int((top - bottom) / resolution) + 1
        print("Output raster shape is ({h},{w})".format(h=height, w=width))

        x_out = np.linspace(left, right, num=width, endpoint=True)
        y_out = np.linspace(bottom, top, num=height, endpoint=True)

        totalAshLoad = np.zeros((height, width), dtype=np.float64)

        for igrain in range(Ngrains):
            thisAshLoad = ash_load_grains[
                row_off[igrain] : row_off[igrain] + h[igrain] - 1,
                col_off[igrain] : col_off[igrain] + w[igrain] - 1,
                igrain,
            ]
            thisAshLoad[thisAshLoad < vmin / Ngrains / 10] = 0
            x = self.x[col_off[igrain] : col_off[igrain] + w[igrain] - 1, igrain]
            y = self.y[row_off[igrain] : row_off[igrain] + h[igrain] - 1, igrain]
            intrp = RectBivariateSpline(y, x, thisAshLoad)

            totalAshLoad += intrp(y_out, x_out)

        totalAshLoad[totalAshLoad < vmin] = nodata
        window = get_data_window(totalAshLoad, nodata=nodata)
        winpad = pad_window(window, height, width)
        totalAshLoad = totalAshLoad[
            winpad["row_start"] : winpad["row_stop"],
            winpad["col_start"] : winpad["col_stop"],
        ]
        x_out = x_out[winpad["col_start"] : winpad["col_stop"]]
        y_out = y_out[winpad["row_start"] : winpad["row_stop"]]

        if export_gtiff:
            self.write_gtiff(
                totalAshLoad,
                x_out,
                y_out,
                export_name,
                nodata=nodata,
                vmin=vmin,
                resolution=None,
            )
        totalAshLoad = np.ma.masked_where(totalAshLoad < vmin, totalAshLoad)

        totalAshLoad, x_out, y_out = _clip_to_window(totalAshLoad, x_out, y_out, vmin=vmin)
        return x_out, y_out, totalAshLoad
    
    def raster_ashload(self,
        resolution=300.0,
        vmin=1e-3,
        nodata=-1,
        export_gtiff=True,
        export_name="AshLoad.tif",
        crs=None):

        Ngrains = self.params.grains.bins

        if Ngrains==1:
            ds = self.raster_ashload_for_grain_class(0, vmin=vmin, nodata=0.0, clipped=True)
        else:
            # Find largest extent, will be smallest grain size
            # imax = np.argmax(self.params.model.xyScale)
            imax = np.argmax(np.maximum(self.params.model.xScale, self.params.model.yScale))

            # Initialize with this ash load
            ds = self.raster_ashload_for_grain_class(imax, vmin= vmin/Ngrains/10, nodata=0.0, clipped=False)

            ds = ds.rio.reproject(ds.rio.crs, resolution=resolution, resampling=Resampling.cubic, nodata=0.0)

            # Loop through remaining, reproject to match
            for igrain in range(Ngrains):
                # skip over the already used result
                if igrain==imax:
                    continue

                ds_tmp = self.raster_ashload_for_grain_class(igrain, vmin= vmin/Ngrains/100, nodata=0.0, clipped=False)

                ds_tmp = ds_tmp.rio.reproject_match(ds, nodata=0.0, resampling=Resampling.cubic)

                ds += ds_tmp

        ds = ds.where(ds.ash_load >= vmin, nodata)

        ds["ash_load"].attrs["short_name"] = "ash_load"
        ds["ash_load"].attrs[
            "long_name"
        ] = "total ash load at ground level for all grain classes"
        ds["ash_load"].attrs["units"] = "kg/m**2"
        ds["ash_load"].rio.write_nodata(nodata, inplace=True)

        ds.x.attrs["units"] = "metres"
        ds.y.attrs["units"] = "metres"

        if crs is not None:
            ds = ds.rio.reproject(crs, resampling=Resampling.cubic)

        return ds

    def contour_settling_flux(self, grain_i, logscale=True, vmin=1e-6):

        self._check_valid_grain(grain_i)
        data = self.SettlingFlux[:, :, grain_i]
        vmax = np.nanmax(data)
        mag = np.log10(vmax)
        if mag > 0:
            mag = ceil(mag)
        else:
            mag = floor(mag)
        vmax = nice_round_up(vmax, mag=10**mag)

        fx = interp1d(np.arange(0, len(self.x)), self.x[:, grain_i] + self.utm[0])
        fy = interp1d(np.arange(0, len(self.y)), self.y[:, grain_i] + self.utm[1])

        if logscale:
            levels = log_levels(vmin, vmax)
        else:
            levels = lin_levels(vmin, vmax)

        g = None
        for level in levels:
            C = find_contours(data, level)

            if len(C) > 0:
                for jj, p in enumerate(C):
                    pxy = np.zeros_like(p)
                    px = fx(p[:, 1])
                    py = fy(p[:, 0])
                    pxy[:, 0] = px
                    pxy[:, 1] = py

                    this_poly = Polygon(pxy)
                    if not this_poly.is_empty:
                        if jj == 0:
                            geom = this_poly
                            g_tmp = gpd.GeoDataFrame(
                                columns=["SettlingFlux", "GrainSize", "geometry"],
                                crs=self.params.source.utmcode,
                            )
                            g_tmp.loc[0, "GrainClass"] = grain_i
                            g_tmp.loc[0, "GrainSize"] = self.params.grains.diameter[
                                grain_i
                            ]
                            g_tmp.loc[0, "SettlingFlux"] = level
                            g_tmp.loc[0, "geometry"] = geom
                        else:
                            if g_tmp.loc[0, "geometry"].contains(this_poly):
                                geom = g_tmp.loc[0, "geometry"].difference(this_poly)
                            else:
                                geom = g_tmp.loc[0, "geometry"].union(this_poly)
                            g_tmp.loc[[0], "geometry"] = gpd.GeoSeries([geom])

            if g is None:
                g = gpd.GeoDataFrame(g_tmp).set_geometry("geometry")
            else:
                g = gpd.GeoDataFrame(concat([g, g_tmp], ignore_index=True))

        g["SettlingFlux"] = g["SettlingFlux"].astype("float64")
        g["GrainClass"] = g["GrainClass"].astype("int64")
        return g

    def plot_settling_flux_for_grain_class(
        self,
        grain_i,
        logscale=True,
        vmin=1e-6,
        vmax=None,
        cmap=plt.cm.Purples,
        basemap=True,
        alpha=0.5,
        max_zoom=None,
        show=True,
        save=False,
        save_name="ashdisperse_result.png",
        min_ax_width=None,
        min_ax_height=None,
    ):
        self._check_valid_grain(grain_i)

        ds = self.raster_settlingflux_for_grain_class(
            grain_i, vmin=vmin, crs=webmerc["init"]
        )
        if ds is None:
            raise RuntimeError("No data in plot_settling_flux_for_grain_class()")

        x = ds.x
        y = ds.y
        data = ds["settling_flux"].values

        x_intrp = interp1d(np.arange(len(x)), x)
        y_intrp = interp1d(np.arange(len(y)), y)

        if vmax is None:
            maxflux = np.nanmax(data)
            mag = np.log10(maxflux)
            if mag > 0:
                mag = ceil(mag)
            else:
                mag = floor(mag)
            vmax = nice_round_up(maxflux, mag=10**mag)
        else:
            mag = np.log10(vmax)
            if mag > 0:
                mag = ceil(mag)
            else:
                mag = floor(mag)

        if vmin is None:
            vmin = 10 ** (mag - 3)

        cbar_fig, cbar_ax = plt.subplots()
        if logscale:
            levels = log_levels(vmin, vmax)
            tmp = cbar_ax.scatter(
                levels, np.ones_like(levels), c=levels, cmap=cmap, norm=LogNorm()
            )
        else:
            levels = lin_levels(vmin, vmax, num=20)
            tmp = cbar_ax.scatter(
                levels, np.ones_like(levels), c=levels, cmap=cmap, norm=Normalize()
            )

        fig, ax = plt.subplots()
        for j, l in enumerate(levels):
            cntrs = find_contours(data, l)
            for c in cntrs:
                ax.fill(
                    x_intrp(c[:, 1]),
                    y_intrp(c[:, 0]),
                    color=cmap(j / len(levels)),
                    alpha=alpha,
                    zorder=1,
                )

        source = self.source_marker.to_crs(webmerc["init"])
        source.plot(ax=ax, marker="^", color="k", markersize=20, zorder=2)

        xlim = list(ax.get_xlim())
        ylim = list(ax.get_ylim())

        x_width = xlim[1] - xlim[0]
        y_height = ylim[1] - ylim[0]

        min_ax_width = int(1.5 * x_width)
        min_ax_height = int(1.5 * y_height)

        set_min_axes(ax, min_width=min_ax_width, min_height=min_ax_height)

        fig = set_figure_size(fig, ax)

        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.1)

        cbar = plt.colorbar(tmp, ax=ax, cax=cax)
        cbar.minorticks_off()
        cbar.set_label("Settling flux (kg/m\u00B2/s)")
        plt.close(cbar_fig)

        ax = ax_ticks(ax, source.geometry[0].x, source.geometry[0].y)
        if basemap:
            ax = add_opentopo_basemap(ax, zorder=0)
            (Narrow, ntext) = add_north_arrow(ax, zorder=11, fontsize=16)
            (scalebar, sbframe) = add_scale_bar(ax, segments=1)

        plt.draw()
        if save:
            plt.savefig(save_name, dpi=300, transparent=True, format="png")

        if show:
            plt.show()
        else:
            plt.close(fig)
        return (fig, ax, cbar)

    def plot_ashload_for_grain_class(
        self,
        grain_i,
        logscale=True,
        vmin=None,
        vmax=10,
        cmap=plt.cm.plasma,
        basemap=True,
        alpha=0.5,
        max_zoom=None,
        show=True,
        save=False,
        save_name="ashdisperse_result.png",
        min_ax_width=None,
        min_ax_height=None,
    ):
        self._check_valid_grain(grain_i)

        ds = self.raster_ashload_for_grain_class(
            grain_i, vmin=1e-6, crs=webmerc["init"]
        )
        if ds is None:
            raise RuntimeError("No data in plot_ashload_for_grain_class()")

        x = ds.x
        y = ds.y
        data = ds["ash_load"].values

        x_intrp = interp1d(np.arange(len(x)), x)
        y_intrp = interp1d(np.arange(len(y)), y)

        if vmax is None:
            maxload = np.nanmax(data)
            mag = np.log10(maxload)
            if mag > 0:
                mag = ceil(mag)
            else:
                mag = floor(mag)
            vmax = nice_round_up(maxload, mag=10**mag)
        else:
            mag = np.log10(vmax)
            if mag > 0:
                mag = ceil(mag)
            else:
                mag = floor(mag)

        if vmin is None:
            vmin = 10 ** (mag - 3)

        cbar_fig, cbar_ax = plt.subplots()
        if logscale:
            levels = log_levels(vmin, vmax)
            tmp = cbar_ax.scatter(
                levels, np.ones_like(levels), c=levels, cmap=cmap, norm=LogNorm()
            )
        else:
            levels = lin_levels(vmin, vmax, num=20)
            tmp = cbar_ax.scatter(
                levels, np.ones_like(levels), c=levels, cmap=cmap, norm=Normalize()
            )

        fig, ax = plt.subplots()
        for j, l in enumerate(levels):
            cntrs = find_contours(data, l)
            for c in cntrs:
                ax.fill(
                    x_intrp(c[:, 1]),
                    y_intrp(c[:, 0]),
                    color=cmap(j / len(levels)),
                    alpha=alpha,
                    zorder=1,
                )

        source = self.source_marker.to_crs(webmerc["init"])
        source.plot(ax=ax, marker="^", color="k", markersize=20, zorder=2)

        xlim = list(ax.get_xlim())
        ylim = list(ax.get_ylim())

        x_width = xlim[1] - xlim[0]
        y_height = ylim[1] - ylim[0]

        min_ax_width = int(1.5 * x_width)
        min_ax_height = int(1.5 * y_height)

        set_min_axes(ax, min_width=min_ax_width, min_height=min_ax_height)

        fig = set_figure_size(fig, ax)

        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.1)

        cbar = plt.colorbar(tmp, ax=ax, cax=cax)
        cbar.minorticks_off()
        # cbar.set_ticks([levels])
        cbar.set_label("Ash load (kg/m\u00B2)")
        plt.close(cbar_fig)

        ax = ax_ticks(ax, source.geometry[0].x, source.geometry[0].y)

        if basemap:
            ax = add_opentopo_basemap(ax, zorder=0)
            (Narrow, ntext) = add_north_arrow(ax, zorder=11, fontsize=16)
            (scalebar, sbframe) = add_scale_bar(ax, segments=1)

        plt.draw()
        if save:
            plt.savefig(save_name, dpi=300, transparent=True, format="png")

        if show:
            plt.show()
        else:
            plt.close(fig)
        return (fig, ax, cbar)

    def contour_ashload_for_grain_class(self, grain_i, cntrs_levels="log", vmin=1e-2):

        self._check_valid_grain(grain_i)

        ashload = self.raster_ashload_for_grain_class(
            grain_i,
            vmin=1e-6,
            nodata=np.nan,
            masked=False,
            crs=webmerc["init"],
        )

        maxload = np.nanmax(ashload["ash_load"])
        mag = np.log10(maxload)
        if mag > 0:
            mag = ceil(mag)
        else:
            mag = floor(mag)
        vmax = nice_round_up(maxload, mag=10**mag)

        if vmin is None:
            vmin = 10 ** (mag - 3)

        if cntrs_levels == "log":
            levels = log_levels(vmin, vmax)
        else:
            levels = lin_levels(vmin, vmax)

        g = self._raster_contour(
            ashload["ash_load"], str(self.params.grains.diameter[grain_i]), levels
        )
        g.rename(columns={"contour": "load", "name": "grain_size"}, inplace=True)
        return g

    def plot_conc_for_grain_class(
        self, grain_i, logscale=True, vmin=1e-6, cmap=plt.cm.bone, basemap=False
    ):
        self._check_valid_grain(grain_i)
        Nz = self.params.output.Nz
        rows, cols = plot_rowscols(Nz + 1)
        fig, axes = plt.subplots(nrows=rows, ncols=cols, sharex=True, sharey=True)
        maxc = np.nanmax(self.Cz[:, :, :, grain_i])
        mag = np.log10(maxc)
        if mag > 0:
            mag = ceil(mag)
        else:
            mag = floor(mag)
        vmax = nice_round_up(maxc, mag=10**mag)

        cbar_fig, cbar_ax = plt.subplots()
        if logscale:
            levels = log_levels(vmin, vmax)
            tmp = cbar_ax.scatter(
                levels, np.ones_like(levels), c=levels, cmap=cmap, norm=LogNorm()
            )
        else:
            levels = lin_levels(vmin, vmax, num=20)
            tmp = cbar_ax.scatter(
                levels, np.ones_like(levels), c=levels, cmap=cmap, norm=Normalize()
            )

        x = self.x[:, grain_i] / 1e3
        y = self.y[:, grain_i] / 1e3
        for j, ax in enumerate(axes.reshape(-1)):
            if j < Nz:
                data = self.Cz[:, :, Nz - j - 1, grain_i]
                if np.nanmax(data) > vmin:
                    # # data = np.ma.masked_where(data <= vmin, data)
                    data[data <= vmin] = np.nan
                    if logscale:
                        CS = ax.contourf(
                            x,
                            y,
                            data,
                            levels,
                            locator=ticker.LogLocator(),
                            cmap=cmap,
                            origin="lower",
                        )
                    else:
                        CS = ax.contourf(
                            x,
                            y,
                            data,
                            levels,
                            cmap=cmap,
                            origin="lower",
                            vmin=vmin,
                            vmax=vmax,
                        )
                ax.set_title(
                    f"z = {self.params.output.altitudes[Nz - j - 1]} m", fontsize=10
                )
            elif j == Nz:
                data = self.C0[:, :, grain_i]
                data[data <= 1e-6] = np.nan
                if np.nanmax(data) > vmin:
                    if logscale:
                        CS = ax.contourf(
                            x,
                            y,
                            data,
                            levels,
                            locator=ticker.LogLocator(),
                            cmap=cmap,
                            origin="lower",
                        )
                    else:
                        CS = ax.contourf(
                            x,
                            y,
                            data,
                            levels,
                            cmap=cmap,
                            origin="lower",
                            vmax=vmax,
                        )
                ax.set_title("z = 0 m")
            else:
                ax.axis("off")

        fig.subplots_adjust(right=0.75)
        cbar_ax = fig.add_axes([0.8, 0.2, 0.05, 0.7])

        cbar = fig.colorbar(tmp, cax=cbar_ax)
        cbar.minorticks_off()
        cbar.set_label("Ash concentration (kg/m\u00B3)")
        plt.close(cbar_fig)

        if basemap:
            ax = add_opentopo_basemap(ax, zorder=0)
            (Narrow, ntext) = add_north_arrow(ax, zorder=11, fontsize=16)
            (scalebar, sbframe) = add_scale_bar(ax, segments=1)

        return (fig, ax, cbar)

    def plot_iso_conc_for_grain_class(self, grain_i, conc):
        self._check_valid_grain(grain_i)
        C = self.Cz[:, :, :, grain_i]

        if np.nanmax(C) < conc:
            message = (
                "In plotIsoConc, no concentration values in excess of"
                + " {}".format(conc)
            )
            print(message)
            return

        # xyScale = self.params.model.xyScale[grain_i]
        xScale = self.params.model.xScale[grain_i]
        yScale = self.params.model.yScale[grain_i]

        z = self.params.output.altitudes

        verts, faces, _, _ = marching_cubes(
            np.where(C > conc, np.ones_like(C), np.zeros_like(C)),
            0.5,
            spacing=(1, 1, 1),
        )

        dx = self.x_dimless[1] - self.x_dimless[0]
        dy = self.y_dimless[1] - self.y_dimless[0]
        dz = z[1] - z[0]

        yv = (
            (verts[:, 0] * dx - np.pi)
            * xScale #xyScale
            * self.params.model.Lx[grain_i]
            / np.pi
            / 1e3
        )
        xv = (
            (verts[:, 1] * dy - np.pi)
            * yScale #xyScale
            * self.params.model.Ly[grain_i]
            / np.pi
            / 1e3
        )
        zv = verts[:, 2] * dz

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")

        ax.plot_trisurf(xv, yv, faces, zv, cmap="Spectral", lw=1)
        ax.set_xlabel("Easting (km)")
        ax.set_ylabel("Northing (km)")
        ax.set_zlabel("Altitude (m)")

    def plot_ashload(
        self,
        resolution=300.0,
        logscale=True,
        cmap=plt.cm.viridis,
        vmin=1e-3,
        nodata=-1,
        alpha=0.5,
        basemap=False,
        export_gtiff=False,
        export_name="AshLoad.tif",
        show=True,
    ):

        ds = self.raster_ashload(
            resolution=resolution,
            vmin=vmin / 10,
            nodata=nodata,
            export_gtiff=export_gtiff,
            export_name=export_name,
            crs=webmerc["init"]
        )
        if ds is None:
            raise RuntimeError("No data in plot_ashload()")

        x = ds.x
        y = ds.y
        data = ds["ash_load"].values

        x_intrp = interp1d(np.arange(len(x)), x)
        y_intrp = interp1d(np.arange(len(y)), y)

        maxc = np.nanmax(data)
        mag = np.log10(maxc)
        if mag > 0:
            mag = ceil(mag)
        else:
            mag = floor(mag)
        vmax = nice_round_up(maxc, mag=10**mag)
        data = np.ma.masked_where(data <= vmin, data)

        cbar_fig, cbar_ax = plt.subplots()
        if logscale:
            levels = log_levels(vmin, vmax)
            tmp = cbar_ax.scatter(
                levels, np.ones_like(levels), c=levels, cmap=cmap, norm=LogNorm()
            )
            # CS = ax.contourf(
            #     x,
            #     y,
            #     data,
            #     levels,
            #     locator=ticker.LogLocator(),
            #     cmap=cmap,
            #     origin="lower",
            # )
        else:
            levels = lin_levels(vmin, vmax, num=20)
            tmp = cbar_ax.scatter(
                levels, np.ones_like(levels), c=levels, cmap=cmap, norm=Normalize()
            )
            # CS = ax.contourf(x, y, data, levels, cmap=cmap, origin="lower")

        fig, ax = plt.subplots()
        for j, l in enumerate(levels):
            cntrs = find_contours(data, l)
            for c in cntrs:
                ax.fill(
                    x_intrp(c[:, 1]),
                    y_intrp(c[:, 0]),
                    color=cmap(j / len(levels)),
                    alpha=alpha,
                    zorder=1,
                )

        source = self.source_marker.to_crs(webmerc["init"])
        source.plot(ax=ax, marker="^", color="k", markersize=20, zorder=2)

        xlim = list(ax.get_xlim())
        ylim = list(ax.get_ylim())

        x_width = xlim[1] - xlim[0]
        y_height = ylim[1] - ylim[0]

        min_ax_width = int(1.5 * x_width)
        min_ax_height = int(1.5 * y_height)

        set_min_axes(ax, min_width=min_ax_width, min_height=min_ax_height)

        fig = set_figure_size(fig, ax)

        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.1)

        cbar = plt.colorbar(tmp, ax=ax, cax=cax)
        cbar.minorticks_off()
        # cbar.set_ticks([levels])
        cbar.set_label("Ash load (kg/m\u00B2)")
        plt.close(cbar_fig)

        ax = ax_ticks(ax, source.geometry[0].x, source.geometry[0].y)

        if basemap:
            ax = add_opentopo_basemap(ax, zorder=0)
            (Narrow, ntext) = add_north_arrow(ax, zorder=11, fontsize=16)
            (scalebar, sbframe) = add_scale_bar(ax, segments=1)

        plt.draw()

        if show:
            plt.show()
        else:
            plt.close(fig)

        return (fig, ax, cbar)
    
    def plot_spectrum_for_grainsize(self, grain_i, vmin=None, vmax=None):
        fig, ax = plt.subplots()
        im = ax.imshow(np.log10(np.absolute(self.C0_FT[:,:,grain_i])), vmin=vmin, vmax=vmax)
        plt.colorbar(im)
        plt.show()

    def folium_ashloads(self, savename, vmin=1e-3):

        m = folium.Map(
            location=[self.params.source.latitude, self.params.source.longitude],
            zoom_start=13,
            control_scale=True,
            prefer_canvas=True,
            tiles="https://{s}.tile.opentopomap.org/{z}/{x}/{y}.png",
            attr="OpenTopoMap"
        )

        style_func = lambda x: {
            "weight": 0.1,
            "color": "black",
            "fillColor": cmap(x["properties"]["load"]),
            "fillOpacity": 0.5,
        }

        for grain_i in range(self.params.grains.bins):
            g = self.contour_ashload_for_grain_class(grain_i, vmin=1e-4)
            g = g.to_crs(webmerc["init"])
            g["geoid"] = g.index.astype(str)
            loads = g[["geoid", "load", "grain_size", "geometry"]]
            loads["grain_size"] = loads["grain_size"].astype(float) * 1e6
            loads["grain_size"] = loads["grain_size"].map("{0:.2f}".format)

            geo_str = loads.to_json()

            if vmin is None:
                vmin = loads.load.min()
            vmax = loads.load.max()

            levels = log_steps(
                nice_round_down(vmin, mag=10 ** np.floor(np.log10(vmin))),
                nice_round_up(vmax),
                step=10,
            )

            cmap = cm.linear.viridis.to_step(
                data=loads["load"], index=levels, method="log", round_method="log10"
            )
            cmap.caption = "ash load (kg/m^2) for {0:.2f} micron grains".format(
                self.params.grains.diameter[grain_i] * 1e6
            )

            lm = folium.features.GeoJson(
                loads,
                style_function=style_func,
                control=True,
                name="Grain size = {0:.2f} microns".format(
                    self.params.grains.diameter[grain_i] * 1e6
                ),
                tooltip=folium.features.GeoJsonTooltip(
                    fields=["grain_size", "load"],
                    aliases=["Grain size (microns)", "Ash load (kg/m^2)"],
                    style=(
                        "background-color: white; color: #333333; font-family: arial; font-size: 12px; padding: 10px;"
                    ),
                    sticky=True,
                ),
            )

            m.add_child(lm)
            m.add_child(cmap)
            m.add_child(BindColormap(lm, cmap))

        folium.LayerControl().add_to(m)

        m.save(savename)



import numpy as np
from numba import (boolean, complex128, float64, int64, njit,
                   parallel_chunksize, prange, set_parallel_chunksize)
from numba.types import Tuple

import ashdisperse.core.core as core
import ashdisperse.spectral.cheb as cheb
from ashdisperse.containers import ChebContainer_type, VelocityContainer_type
from ashdisperse.met.met import MetData_type
from ashdisperse.params import Parameters_type

# @njit(
#     Tuple((complex128[::1], complex128[:, ::1]))(
#         float64, float64, complex128[::1], Parameters_type, MetData_type
#     ),
#     cache=True,
#     parallel=False,
#     fastmath=True,
# )
# def ade_mode_rk_solve(kx, ky, fxy_ij, parameters, Met):

#     z = np.zeros((parameters.output.Nz), dtype=np.float64)
#     z = parameters.output.altitudes
#     z = z / parameters.source.PlumeHeight

#     conc_0_ft = np.zeros((parameters.grains.bins), dtype=np.complex128)
#     conc_z_ft = np.zeros((z.size, parameters.grains.bins), dtype=np.complex128)

#     for grain_i in range(parameters.grains.bins):
#         C = core.rkBVP(kx, ky, fxy_ij[grain_i], grain_i, parameters, Met)

#         conc_0_ft[grain_i] = C[-1]
#         conc_z_ft[:, grain_i] = C[::-1]

#     return conc_0_ft, conc_z_ft


@njit(
    Tuple((complex128, complex128[::1]))(
        int64,
        float64,
        float64,
        complex128,
        ChebContainer_type,
        Parameters_type,
        VelocityContainer_type,
    ),
    cache=True,
    parallel=False,
    fastmath=True,
)
def ade_mode_solve_grain(grain_i, kx, ky, fxy_ij, ChebContainer, parameters, VelocityContainer):

    z = np.zeros((parameters.output.Nz), dtype=np.float64)
    z = parameters.output.altitudes
    z = z / parameters.source.PlumeHeight
    Nlow = z[z <= 1].size

    conc_0_ft = np.complex128(0.0)
    conc_z_ft = np.zeros((z.size), dtype=np.complex128)

    """Lower domain.
    Solve for:
      inhomogeneous part (called p1) with null boundary conditions
      homogeneous part (called r1) with bcs r1(1) = 1, r1(-1) = 0
      homogeneous part (called l1) with bcs l1(1) = 0, l1(-1) = 1
      with _c for coefficients, _v for values)
    """
    p1_c, l1_c, r1_c = core.LowerODE(
        kx,
        ky,
        fxy_ij,
        grain_i,
        ChebContainer,
        parameters,
        VelocityContainer,
    )

    dp1_p1 = cheb.cheb_dif_p1(p1_c)
    dr1_p1 = cheb.cheb_dif_p1(r1_c)
    dl1_p1 = cheb.cheb_dif_p1(l1_c)

    dp1_m1 = cheb.cheb_dif_m1(p1_c)
    dr1_m1 = cheb.cheb_dif_m1(r1_c)
    dl1_m1 = cheb.cheb_dif_m1(l1_c)

    """Upper domain.
         Solve for:
           inhomogeneous part (called p2) with null boundary conditions
           homogeneous part (called r2) with bcs r2(1) = 1, r2(-1) = 0
           with _c for coefficients, _v for values)
    """
    l2_c = core.UpperODE(
        kx, ky, grain_i, ChebContainer, parameters, VelocityContainer
    )

    dl2_m1 = cheb.cheb_dif_m1(l2_c)

    a = dl1_m1
    b = dr1_m1
    c = dl1_p1
    d = dr1_p1 - dl2_m1

    b0 = -dp1_m1
    b1 = -dp1_p1

    delta = a * d - b * c

    match0 = (b0 * d - b1 * b) / delta
    match1 = (b1 * a - b0 * c) / delta

    Cheb_r1 = np.zeros(
        (parameters.output.Cheb_lower.shape[0], r1_c.size), dtype=np.complex128
    )
    Cheb_l1 = np.zeros(
        (parameters.output.Cheb_lower.shape[0], l1_c.size), dtype=np.complex128
    )
    Cheb_p1 = np.zeros(
        (parameters.output.Cheb_lower.shape[0], p1_c.size), dtype=np.complex128
    )

    Cheb_r1[:, : r1_c.size] = parameters.output.Cheb_lower[:, : r1_c.size]
    Cheb_l1[:, : l1_c.size] = parameters.output.Cheb_lower[:, : l1_c.size]
    Cheb_p1[:, : p1_c.size] = parameters.output.Cheb_lower[:, : p1_c.size]

    r1_v = Cheb_r1 @ r1_c
    l1_v = Cheb_l1 @ l1_c
    p1_v = Cheb_p1 @ p1_c

    conc_z_ft[:Nlow] = p1_v + match0 * l1_v + match1 * r1_v

    Cheb_l2 = np.zeros(
        (parameters.output.Cheb_upper.shape[0], l2_c.size), dtype=np.complex128
    )
    Cheb_l2[:, : l2_c.size] = parameters.output.Cheb_upper[:, : l2_c.size]
    l2_v = Cheb_l2 @ l2_c

    conc_z_ft[Nlow:] = match1 * l2_v

    conc_0_ft = match0  # (b0*d - b1*b)/delta

    return conc_0_ft, conc_z_ft


@njit(
    Tuple((complex128[::1], complex128[:, ::1]))(
        float64,
        float64,
        complex128[::1],
        ChebContainer_type,
        Parameters_type,
        VelocityContainer_type,
    ),
    cache=True,
    parallel=False,
    fastmath=True,
)
def ade_mode_solve(kx, ky, fxy_ij, ChebContainer, parameters, VelocityContainer):

    z = np.zeros((parameters.output.Nz), dtype=np.float64)
    z = parameters.output.altitudes
    z = z / parameters.source.PlumeHeight
    Nlow = z[z <= 1].size

    conc_0_ft = np.zeros((parameters.grains.bins), dtype=np.complex128)
    conc_z_ft = np.zeros((z.size, parameters.grains.bins), dtype=np.complex128)

    for grain_i in range(parameters.grains.bins):
        """Lower domain.
        Solve for:
          inhomogeneous part (called p1) with null boundary conditions
          homogeneous part (called r1) with bcs r1(1) = 1, r1(-1) = 0
          homogeneous part (called l1) with bcs l1(1) = 0, l1(-1) = 1
          with _c for coefficients, _v for values)
        """
        p1_c, l1_c, r1_c = core.LowerODE(
            kx,
            ky,
            fxy_ij[grain_i],
            np.int64(grain_i),
            ChebContainer,
            parameters,
            VelocityContainer,
        )

        dp1_p1 = cheb.cheb_dif_p1(p1_c)
        dr1_p1 = cheb.cheb_dif_p1(r1_c)
        dl1_p1 = cheb.cheb_dif_p1(l1_c)

        dp1_m1 = cheb.cheb_dif_m1(p1_c)
        dr1_m1 = cheb.cheb_dif_m1(r1_c)
        dl1_m1 = cheb.cheb_dif_m1(l1_c)

        """Upper domain.
             Solve for:
               inhomogeneous part (called p2) with null boundary conditions
               homogeneous part (called r2) with bcs r2(1) = 1, r2(-1) = 0
               with _c for coefficients, _v for values)
        """
        l2_c = core.UpperODE(
            kx, ky, np.int64(grain_i), ChebContainer, parameters, VelocityContainer
        )

        dl2_m1 = cheb.cheb_dif_m1(l2_c)

        a = dl1_m1
        b = dr1_m1
        c = dl1_p1
        d = dr1_p1 - dl2_m1

        b0 = -dp1_m1
        b1 = -dp1_p1

        delta = a * d - b * c

        match0 = (b0 * d - b1 * b) / delta
        match1 = (b1 * a - b0 * c) / delta

        Cheb_r1 = np.zeros(
            (parameters.output.Cheb_lower.shape[0], r1_c.size), dtype=np.complex128
        )
        Cheb_l1 = np.zeros(
            (parameters.output.Cheb_lower.shape[0], l1_c.size), dtype=np.complex128
        )
        Cheb_p1 = np.zeros(
            (parameters.output.Cheb_lower.shape[0], p1_c.size), dtype=np.complex128
        )

        Cheb_r1[:, : r1_c.size] = parameters.output.Cheb_lower[:, : r1_c.size]
        Cheb_l1[:, : l1_c.size] = parameters.output.Cheb_lower[:, : l1_c.size]
        Cheb_p1[:, : p1_c.size] = parameters.output.Cheb_lower[:, : p1_c.size]

        r1_v = Cheb_r1 @ r1_c
        l1_v = Cheb_l1 @ l1_c
        p1_v = Cheb_p1 @ p1_c

        conc_z_ft[:Nlow, grain_i] = p1_v + match0 * l1_v + match1 * r1_v

        Cheb_l2 = np.zeros(
            (parameters.output.Cheb_upper.shape[0], l2_c.size), dtype=np.complex128
        )
        Cheb_l2[:, : l2_c.size] = parameters.output.Cheb_upper[:, : l2_c.size]
        l2_v = Cheb_l2 @ l2_c

        conc_z_ft[Nlow:, grain_i] = match1 * l2_v

        conc_0_ft[grain_i] = match0  # (b0*d - b1*b)/delta

    return conc_0_ft, conc_z_ft


@njit(
    Tuple((complex128[:, :, ::1], complex128[:, :, :, ::1]))(
        float64[::1],
        float64[::1],
        complex128[:, :, ::1],
        ChebContainer_type,
        Parameters_type,
        VelocityContainer_type,
    ),
    parallel=True,
    cache=True,
    fastmath=False,
)
def ade_ft_system(kx, ky, fxy_f, cheby, params, velocities):

    Nx = kx.size
    Ny = ky.size

    Ng = params.grains.bins
    Nz = params.output.Nz

    fft_tol = params.solver.fft_tol

    conc_0_fft = np.zeros((Ny, Nx//2+1, Ng), dtype=np.complex128)
    conc_z_fft = np.zeros(
        (Ny, Nx//2+1, Nz, Ng), dtype=np.complex128
    )

    Xstop = (Nx//2+1) * np.ones((Ng), dtype=np.int64)
    Ystop = Ny//2 * np.ones((Ng), dtype=np.int64)

    half_Ny = (Ny+1)//2

    N = (Nx//2 + 1) * (half_Ny) * Ng
    for idx in prange(N):
        i = idx // (half_Ny * Ng)  # Compute i
        j = (idx // Ng) % (half_Ny)  # Compute j
        j_upper = Ny - j - 1  # Mirror j
        #k = idx % Ng          # Compute k
        k = Ng - 1 - (idx % Ng) # Compute k in reverse

        if (i<=Xstop[k] and j<=Ystop[k]):

            conc_0_mode_fft, conc_z_mode_fft = ade_mode_solve_grain(k,
                    kx[i], ky[j], fxy_f[j, i, k], cheby, params, velocities
                )
            
            conc_0_fft[j, i, k] = conc_0_mode_fft
            conc_z_fft[j, i, :, k] = conc_z_mode_fft
            
            conc_0_mode_fft_upper, conc_z_mode_fft_upper = ade_mode_solve_grain(k,
                    kx[i], ky[j_upper], fxy_f[j_upper, i, k], cheby, params, velocities
                )
        
            conc_0_fft[j_upper, i, k] = conc_0_mode_fft_upper
            conc_z_fft[j_upper, i, :, k] = conc_z_mode_fft_upper

            if i==0:
                if np.absolute(conc_0_mode_fft)< fft_tol and np.absolute(conc_0_mode_fft_upper) < fft_tol:
                    Ystop[k] = np.minimum(j, Ystop[k])
            if j==0:
                if 0 < np.absolute(conc_0_mode_fft) < fft_tol:
                    Xstop[k] = np.minimum(i, Xstop[k])
        else:
            conc_0_fft[j, i, k] = np.complex128(0.0)
            conc_z_fft[j, i, :, k] = np.zeros((Nz), dtype=np.complex128)
            conc_0_fft[j_upper, i, k] = np.complex128(0.0)
            conc_z_fft[j_upper, i, :, k] = np.zeros((Nz), dtype=np.complex128)
    
    return conc_0_fft, conc_z_fft


@njit(
    Tuple((complex128[:, :, ::1], complex128[:, :, :, ::1]))(
        complex128[:, :, ::1],
        complex128[:, :, :, ::1],
        float64[::1],
        float64[::1],
        complex128[:, :, ::1],
        ChebContainer_type,
        Parameters_type,
        VelocityContainer_type,
        boolean,
    ),
    parallel=True,
    cache=True,
    fastmath=True,
)
def ade_ft_refine(conc_0_fft_old, conc_z_fft_old, kx, ky, fxy_f, cheby, params, velocities, full=False):

    Ny_old = conc_0_fft_old.shape[0]
    half_Nx_old = conc_0_fft_old.shape[1] - 1

    Nx_old = half_Nx_old * 2
    half_Ny_old = Ny_old // 2

    Nx = kx.size
    Ny = ky.size

    half_Nx = Nx // 2 # Only need the non-negative modes but symmetry of FFT of real-valued function
    half_Ny = (Ny + 1) // 2

    Ng = params.grains.bins
    Nz = params.output.Nz

    fft_tol = params.solver.fft_tol

    Xstop = (Nx//2+1) * np.ones((Ng), dtype=np.int64)
    Ystop = Ny//2 * np.ones((Ng), dtype=np.int64)

    conc_0_fft = np.zeros((Ny, half_Nx+1, Ng), dtype=np.complex128)
    conc_z_fft = np.zeros((Ny, Nx+1, Nz, Ng), dtype=np.complex128)

    N = (Nx//2 + 1) * (half_Ny) * Ng
    for idx in prange(N):
        i = idx // (half_Ny * Ng)  # Compute i
        j = (idx // Ng) % (half_Ny)  # Compute j
        j_upper = Ny - j - 1  # Mirror j
        k = idx % Ng          # Compute k

        j_upper = Ny - j - 1 # Mirror j
        j_upper_old = Ny_old - j - 1 # Mirror j

        is_special = (i == 0) or (j == 0)

        if full:
            if (not is_special and i<half_Nx_old and j<half_Ny_old):
            
                if np.absolute(conc_0_fft_old[j, i, k])<fft_tol:
                    conc_0_mode_fft, conc_z_mode_fft = ade_mode_solve_grain(k,
                        kx[i], ky[j], fxy_f[j, i, k], cheby, params, velocities
                    )
                else:
                    conc_0_mode_fft = conc_0_fft_old[j, i, k]
                    conc_z_mode_fft = conc_z_fft_old[j, i, :, k]

                if np.absolute(conc_0_fft_old[j, i, k])<fft_tol:
                    conc_0_mode_fft_upper, conc_z_mode_fft_upper = ade_mode_solve_grain(k,
                        kx[i], ky[j_upper], fxy_f[j_upper, i, k], cheby, params, velocities
                    )
                else:
                    conc_0_mode_fft_upper = conc_0_fft_old[j_upper_old, i, k]
                    conc_z_mode_fft_upper = conc_z_fft_old[j_upper_old, i, :, k]
        
            else:
                conc_0_mode_fft, conc_z_mode_fft = ade_mode_solve_grain(k,
                    kx[i], ky[j], fxy_f[j, i, k], cheby, params, velocities
                )
                
                conc_0_mode_fft_upper, conc_z_mode_fft_upper = ade_mode_solve_grain(k,
                    kx[i], ky[j_upper], fxy_f[j_upper, i, k], cheby, params, velocities
                )
            
            conc_0_fft[j, i, k] = conc_0_mode_fft
            conc_z_fft[j, i, :, k] = conc_z_mode_fft

            conc_0_fft[j_upper, i, k] = conc_0_mode_fft_upper
            conc_z_fft[j_upper, i, :, k] = conc_z_mode_fft_upper
        
        else:
            if (not is_special and i<half_Nx_old and j<half_Ny_old):
                
                conc_0_mode_fft = conc_0_fft_old[j, i, k]
                conc_z_mode_fft = conc_z_fft_old[j, i, :, k]

                conc_0_mode_fft_upper = conc_0_fft_old[j_upper_old, i, k]
                conc_z_mode_fft_upper = conc_z_fft_old[j_upper_old, i, :, k]

            else:
                if (i<=Xstop[k] and j<=Ystop[k]):

                    conc_0_mode_fft, conc_z_mode_fft = ade_mode_solve_grain(k,
                        kx[i], ky[j], fxy_f[j, i, k], cheby, params, velocities
                    )

                    conc_0_mode_fft_upper, conc_z_mode_fft_upper = ade_mode_solve_grain(k,
                        kx[i], ky[j_upper], fxy_f[j_upper, i, k], cheby, params, velocities
                    )

                    if i==0:
                        if np.absolute(conc_0_mode_fft)< fft_tol and np.absolute(conc_0_mode_fft_upper) < fft_tol:
                            Ystop[k] = np.minimum(j, Ystop[k])
                    if j==0:
                        if 0 < np.absolute(conc_0_mode_fft) < fft_tol:
                            Xstop[k] = np.minimum(i, Xstop[k])
                
                else:
                    conc_0_mode_fft = np.complex128(0.0)
                    conc_z_mode_fft = np.zeros((Nz), dtype=np.complex128)
                    conc_0_mode_fft_upper = np.complex128(0.0)
                    conc_z_mode_fft_upper = np.zeros((Nz), dtype=np.complex128)
            
            conc_0_fft[j, i, k] = conc_0_mode_fft
            conc_z_fft[j, i, :, k] = conc_z_mode_fft

            conc_0_fft[j_upper, i, k] = conc_0_mode_fft_upper
            conc_z_fft[j_upper, i, :, k] = conc_z_mode_fft_upper
            
    return conc_0_fft, conc_z_fft

# @njit(
#     Tuple((complex128[:, :, ::1], complex128[:, :, :, ::1]))(
#         float64[::1],
#         float64[::1],
#         complex128[:, :, ::1],
#         Parameters_type,
#         MetData_type,
#     ),
#     parallel=True,
#     cache=True,
#     fastmath=True,
# )
# def ade_ft_system_rk(kx, ky, fxy_f, params, Met):

#     Nx = kx.size
#     Ny = ky.size

#     conc_0_fft = np.zeros((Ny, Nx, params.grains.bins), dtype=np.complex128)
#     conc_z_fft = np.zeros(
#         (Ny, Nx, params.output.Nz, params.grains.bins), dtype=np.complex128
#     )

#     # do kx = ky = 0:
#     conc_0_mode_fft, conc_z_mode_fft = ade_mode_rk_solve(
#         0.0, 0.0, fxy_f[0, 0, :], params, Met
#     )
#     conc_0_fft[0, 0, :] = conc_0_mode_fft
#     conc_z_fft[0, 0, :, :] = conc_z_mode_fft

#     # do kx = 0, ky = 1 ... Ny/2-1, -Ny/2
#     # and we get kx = 0, ky = -Ny/2+1, ... , -1 for free by conjugation
#     for j in prange(1, Ny // 2 + 1):

#         conc_0_mode_fft, conc_z_mode_fft = ade_mode_rk_solve(
#             0.0, ky[j], fxy_f[j, 0, :], params, Met
#         )

#         conc_0_fft[j, 0, :] = conc_0_mode_fft
#         conc_0_fft[Ny - j, 0, :] = np.conj(conc_0_mode_fft)

#         conc_z_fft[j, 0, :, :] = conc_z_mode_fft
#         conc_z_fft[Ny - j, 0, :, :] = np.conj(conc_z_mode_fft)

#     # do ky = 0, kx = 1 ... Nx/2-1, -Nx/2
#     # and we get ky = 0, kx = -Nx/2+1, ... , -1 for free by conjugation
#     for i in prange(1, Nx // 2 + 1):

#         conc_0_mode_fft, conc_z_mode_fft = ade_mode_rk_solve(
#             kx[i], 0.0, fxy_f[0, i, :], params, Met
#         )

#         conc_0_fft[0, i, :] = conc_0_mode_fft
#         conc_0_fft[0, Nx - i, :] = np.conj(conc_0_mode_fft)

#         conc_z_fft[0, i, :, :] = conc_z_mode_fft
#         conc_z_fft[0, Nx - i, :, :] = np.conj(conc_z_mode_fft)

#     # Do first quadrant, kx = 1 ... Nx/2-1, ky = 1 .. Ny/2-1
#     # and get the 4th quadrant for free by conjugation.
#     # Also do second quadrant, kx = -Nx/2+1 ... -1, ky = 1 .. Ny/2-1
#     # and get the third quadrant for free by conjugation.
#     for i in prange(1, Nx // 2):
#         for j in range(1, Ny // 2):

#             conc_0_mode_fft, conc_z_mode_fft = ade_mode_rk_solve(
#                 kx[i], ky[j], fxy_f[j, i, :], params, Met
#             )
#             conc_0_fft[j, i, :] = conc_0_mode_fft
#             conc_z_fft[j, i, :, :] = conc_z_mode_fft

#             conc_0_mode_fft, conc_z_mode_fft = ade_mode_rk_solve(
#                 kx[Nx - i], ky[j], fxy_f[j, Nx - i, :], params, Met
#             )
#             conc_0_fft[j, Nx - i, :] = conc_0_mode_fft
#             conc_z_fft[j, Nx - i, :, :] = conc_z_mode_fft

#             conc_0_fft[Ny - j, Nx - i, :] = np.conj(conc_0_fft[j, i, :])
#             conc_0_fft[Ny - j, i, :] = np.conj(conc_0_fft[j, Nx - i, :])

#             conc_z_fft[Ny - j, Nx - i, :, :] = np.conj(conc_z_fft[j, i, :, :])
#             conc_z_fft[Ny - j, i, :, :] = np.conj(conc_z_fft[j, Nx - i, :, :])

#     return conc_0_fft, conc_z_fft
#     return conc_0_fft, conc_z_fft



from pkg_resources import resource_filename

from ashdisperse.ashdisperse import load_result, refine, setup, solve
from ashdisperse.core import AshDisperseResult
from ashdisperse.interface import (advected_settling_trajectories,
                                   estimate_dispersal_distance, load_inputs,
                                   save_inputs, set_met, set_parameters)
from ashdisperse.met import load_met, save_met, wind_plot
from ashdisperse.params import EmissionParameters as _EmissionParameters
from ashdisperse.params import GrainParameters as _GrainParameters
from ashdisperse.params import OutputParameters as _OutputParameters
from ashdisperse.params import Parameters as _Parameters
from ashdisperse.params import PhysicalParameters as _PhysicalParameters
from ashdisperse.params import SolverParameters as _SolverParameters
from ashdisperse.params import SourceParameters as _SourceParameters
from ashdisperse.params import (copy_parameters, load_parameters,
                                save_parameters, update_parameters)
from ashdisperse.version import __version__

__author__ = "Mark J. Woodhouse"

def __initialize__():

    params = _Parameters()
    params.source = _SourceParameters(
        51.456255,
        -2.604762,
        32632,
        10000,
        10000,
        1e6,
        18000,
        name='initialize',
    )
    params.grains = _GrainParameters()
    params.grains.add_grain(1e-3, 1200, 1)
    params.emission = _EmissionParameters()
    params.emission.add_profile(0, 10000, 0, 10)
    params.solver = _SolverParameters(
        domX=1.5,
        domY=1.5,
        minN_log2=4,
        maxN_log2=8,
        epsilon=1e-8,
        Nx_log2=5,
        Ny_log2=5,
    )
    params.physical = _PhysicalParameters()
    params.output = _OutputParameters(0, 10000, 1000)
    params.output.set_altitudes()
    params.output.ChebMats(params.solver.maxN, params.source.PlumeHeight)

    metdata = resource_filename(__name__, "/data/metdata.npz")

    met = load_met(metdata)

    return params, met


from collections import OrderedDict
from typing import Optional

import numpy as np
from numba import float64, int64, optional
from numba.experimental import jitclass
from numba.types import unicode_type

from ashdisperse.utilities import latlon_point_to_utm_code

source_spec = OrderedDict()
source_spec["latitude"] = float64  # latitude of source
source_spec["longitude"] = float64  # longitude of source
source_spec["utmcode"] = int64  # EPSG utm code of source
source_spec["radius"] = float64  # radius of the source
source_spec["PlumeHeight"] = float64  # height of the source plume
source_spec["MER"] = float64  # mass eruption rate
source_spec["duration"] = float64  # duration of eruption
source_spec["name"] = unicode_type  # name of source


@jitclass(source_spec)
class SourceParameters:
    def __init__(self,
                 latitude:np.float64=np.nan,
                 longitude:np.float64=np.nan,
                 utmcode:int=0,
                 radius:np.float64=np.nan,
                 PlumeHeight:np.float64=np.nan,
                 MER:np.float64=np.nan,
                 duration:np.float64=np.nan,
                 name:str=""):
        self.latitude = latitude
        self.longitude = longitude
        self.utmcode = utmcode
        self.radius = radius
        self.PlumeHeight = PlumeHeight
        self.MER = MER
        self.duration = duration
        self.name = name

    def validate(self):
        if np.isnan(self.latitude) or np.abs(self.latitude)>90:
            raise ValueError("In SourceParameters, latitude must be in the range [-90, 90]")
        if np.isnan(self.longitude) or np.abs(self.longitude)>180:
            raise ValueError("In SourceParameters, longitude must be in the range [-180, 180]")
        if np.isnan(self.radius) or self.radius < 0:
            raise ValueError("In SourceParameters, radius must be positive")
        if np.isnan(self.PlumeHeight) or self.PlumeHeight < 0:
            raise ValueError("In SourceParameters, PlumeHeight must be positive")
        if np.isnan(self.MER) or self.MER < 0:
            raise ValueError("In SourceParameters, MER must be positive")
        if np.isnan(self.duration) or self.duration < 0:
            raise ValueError("In SourceParameters, duration must be positive")
        return 1

    def utm_from_latlon(self):
        offset = int(np.round((183+self.longitude)/6.0))
        self.utmcode = int(32600+offset) if (self.latitude > 0) else int(32700+offset)
        return

    def from_values(self,
        lat,
        lon,
        utmcode=None,
        radius=10e3,
        PlumeHeight=10e3,
        MER=1e6,
        duration=18000,
        name="",
    ):

        self.latitude = np.float64(lat)
        self.longitude = np.float64(lon)
        if utmcode is None:
            self.utm_from_latlon()
        else:
            self.utmcode = utmcode

        self.radius = np.float64(radius)

        self.PlumeHeight = np.float64(PlumeHeight)

        self.MER = np.float64(MER)

        self.duration = np.float64(duration)

        self.name = name

        self.validate()

    def describe(self):
        print("Source parameters for AshDisperse")
        print("  Mass eruption rate MER = ", self.MER, " kg/s")
        print("  Eruption duration = ", self.duration, " s")
        print("  Plume height H = ", self.PlumeHeight, " m")
        print("  Gaussian source radius = ", self.radius, " m")
        print("********************")

# pylint: disable=E1101
SourceParameters_type = SourceParameters.class_type.instance_type

def _source_dict(p: SourceParameters):
    return {
        "name": p.name,
        "latitude": p.latitude,
        "longitude": p.longitude,
        "utmcode": p.utmcode,
        "radius": p.radius,
        "PlumeHeight": p.PlumeHeight,
        "MER": p.MER,
        "duration": p.duration,
    }


def _source_params_equal(p1: SourceParameters, p2: SourceParameters) -> bool:

    test = (
        (p1.name == p2.name) and
        (p1.latitude == p2.latitude) and
        (p1.longitude == p2.longitude) and
        (p1.utmcode == p2.utmcode) and
        (p1.radius == p2.radius) and
        (p1.PlumeHeight == p2.PlumeHeight) and
        (p1.MER == p2.MER) and
        (p1.duration == p2.duration)
    )
    return test


from .emission_params import (EmissionParameters, EmissionParameters_type,
                              Suzuki_k_from_peak, Suzuki_peak_from_k,
                              _emission_dict)
from .grain_params import GrainParameters, GrainParameters_type, _grains_dict
from .met_params import MetParameters, MetParameters_type
from .model_params import ModelParameters, ModelParameters_type
from .output_params import OutputParameters, OutputParameters_type
from .params import (Parameters, Parameters_type, copy_parameters,
                     load_parameters, save_parameters, update_parameters)
from .physical_params import (PhysicalParameters, PhysicalParameters_type,
                              _physical_dict)
from .solver_params import (SolverParameters, SolverParameters_type,
                            _solver_dict)
from .source_params import (SourceParameters, SourceParameters_type,
                            _source_dict)




from .datetime_strings import is_valid_datetime, string_to_datetime
from .geo import latlon_point_to_utm_code, webmerc, wgs84
from .number_utils import (lin_levels, log_levels, log_steps, nice_round_down,
                           nice_round_up)
from .plotting import pad_window, plot_rowscols



# -*- coding: utf-8 -*-
"""Emission profile parameters for AshDisperse.

Defines the EmissionParameters class, containing attributes specifying emission
classes and methods for amending these, and a printer for this object.

There can be a single emission profile, used for all grain classes,
or a different emission profile for each grain class

Example:
    emission = EmissionParameters()
    
    print_emission_parameters(emission)
"""

from collections import OrderedDict

import numpy as np
from numba import float64, int64
from numba.experimental import jitclass
from numba.typed import List
from numba.types import ListType

emission_spec = OrderedDict()
emission_spec["lower"] = ListType(float64)  # Minimum altitude for release
emission_spec["upper"] = ListType(float64)  # Maximum altitude for release
emission_spec["profile"] = ListType(int64)  # profile type enumerator: =0 for suzuki, =1 for uniform
emission_spec["Suzuki_k"] = ListType(float64)  # Suzuki_k parameter
"""OrderedDict: EmissionParameters attribute type specification.

This specification is required for numba jitting.
"""

@jitclass(emission_spec)
class EmissionParameters:
    """Emission profile parameters required for AshDisperse.

    Defines the EmissionParameters class, containing attributes specifying emission
    classes and methods for amending these.

    Attributes:
        lower: A list of floats containing the minimum altitude for release of each grain class.
        Suzuki_k: A list of floats containing the Suzuki k parameter of each grain class.
    """

    def __init__(self):
        """Initialize of EmissionParameters to empty lists.

        Use add_profile to add profile for grain class.
        """
        self.lower = List.empty_list(np.float64)
        self.upper = List.empty_list(np.float64)
        self.profile = List.empty_list(np.int64)
        self.Suzuki_k = List.empty_list(np.float64)
    
    def len(self):
        return len(self.lower)
    
    def validate(self):
        """Validate grain classes in GrainParameters object."""
        if (self.len()==0):
            return 0
        if len(self.upper) != self.len():
            raise ValueError(
                "In EmissionParameters: number of upper values must equal number of lower values"
            )
        if len(self.profile) != self.len():
            raise ValueError(
                "In EmissionParameters: number of profile values must equal number of lower values"
            )
        if len(self.Suzuki_k) != self.len():
            raise ValueError(
                "In EmissionParameters: number of Suzuki_k values must equal number of lower values"
            )
        
        for j in range(self.len()):
            if (self.lower[j]<0):
                raise ValueError("In EmissionParameters: lower values must be non-negative")
            if (self.upper[j]<0):
                raise ValueError("In EmissionParameters: upper values must be non-negative")
            if (self.upper[j] - self.lower[j]<0):
                raise ValueError("In EmissionParameters: upper values must be larger than lower values")
            if (self.Suzuki_k[j]<=0):
                raise ValueError("In EmissionParameters: Suzuki_k values must be positive")
        return 1

    def from_values(self, lower, upper, profile, Suzuki_k):
        self.add_profile(lower, upper, profile, Suzuki_k)
        return

    def from_lists(self, lower, upper, profile, Suzuki_k):
        """Initialize emission parameters from lists"""

        if len(lower) != len(profile):
            raise ValueError("Size of lower and profile not equal")
        
        if len(lower) != len(upper):
            raise ValueError("Size of lower and upper not equal")
        
        if len(lower) != len(Suzuki_k):
            raise ValueError("Size of lower and Suzuki_k not equal")

        self.clear()

        for j, (l, u, p, k) in enumerate(zip(lower, upper, profile, Suzuki_k)):
            self.add_profile(l, u, p, k)

        return
    
    def clear(self):
        self.lower = List.empty_list(np.float64)
        self.upper = List.empty_list(np.float64)
        self.profile = List.empty_list(np.int64)
        self.Suzuki_k = List.empty_list(np.float64)

    def add_profile(self, lower, upper, profile, Suzuki_k):
        """Add a emission class to the EmissionParameters object.

        An emission profile requires a lower altitude and a Suzuki k parameter.

        Args:
            lower: [float64] The lower altitude of the emission profile;
                must be positive.
            upper: [float64] The upper altitude of the emission profile;
                must be positive and > lower.
            profile: [int64] The type of emission profile: =0 for Suzuki, =1 for uniform
                must be either 0 or 1
            Suzuki_k: [float64] The Suzuki k parameter of the emission profile;
                must be positive.
        """
        if lower < 0:
            raise ValueError("Emission profile lower altitude must be positive")
        if upper < 0:
            raise ValueError("Emission profile upper altitude must be positive")
        if upper < lower:
            raise ValueError("Emission profile must have upper altitude > lower altitude")
        if profile not in [0,1]:
            raise ValueError("Emission profile must have profile type of either 0 or 1")
        if Suzuki_k < 0:
            raise ValueError("Emission profile Suzukik parameter must be positive")
        self.lower.append(np.float64(lower))
        self.upper.append(np.float64(upper))
        self.profile.append(np.int64(profile))
        self.Suzuki_k.append(np.float64(Suzuki_k))

    def remove_profile(self, profile_i):
        """Remove a emission profile from the EmissionParameters object.

        The emission profile with a given index is removed from the EmissionParameters
        object.
        
        Args:
            profile_i: [int] The index of the emission profile.  Must be less than the
                           number of profiles.
        """
        if profile_i < 0:
            raise ValueError("index to profile class must be positive")
        if profile_i >= len(self.lower):
            raise ValueError(f"index must in range [0,{self.len()-1}]")
        self.lower.pop(profile_i)
        self.upper.pop(profile_i)
        self.profile.pop(profile_i)
        self.Suzuki_k.pop(profile_i)

    def describe(self):
        """Describe the emission profiles in EmissionParameters object."""
        print("Emission profile parameters for AshDisperse")
        for j in range(len(self.lower)):
            print(f"  Emission profile for grain class {j+1}")
            print("    Lower altitude =", self.lower[j], " m")
            print("    Upper altitude =", self.upper[j], " m")
            if self.profile[j]==0:
                print(f"    Profile type = 0 (Suzuki)")
            else:
                print(f"    Profile type = 1 (Uniform)")
            print("    Suzuki k = ", self.Suzuki_k[j])
            # print("    Suzuki peak = ", Suzuki_peak_from_k(self.Suzuki_k[j], self.lower[j], self.upper[j]), " m")
        print("********************")

# pylint: disable=E1101
EmissionParameters_type = EmissionParameters.class_type.instance_type

def Suzuki_k_from_peak(Suzuki_peak, lower, upper):
    Suzuki_k = (upper-lower)/(upper-Suzuki_peak)
    return Suzuki_k

def Suzuki_peak_from_k(Suzuki_k, lower, upper):
    Suzuki_peak = upper - (upper-lower)/Suzuki_k
    return Suzuki_peak

def _emission_dict(p):
    return {
        'lower': list(p.lower),
        'upper': list(p.upper),
        'profile': list(p.profile),
        'Suzuki_k': list(p.Suzuki_k),
    }

def _emission_params_equal(p1: EmissionParameters, p2: EmissionParameters) -> bool:
    n = len(p1.lower)
    if not (n == len(p2.lower)):
        return False
    for i in range(n):
        if (p1.lower[i] != p2.lower[i]):
            return False
        if (p1.upper[i] != p2.upper[i]):
            return False
        if (p1.profile[i] != p2.profile[i]):
            return False
        if (p1.Suzuki_k[i] != p2.Suzuki_k[i]):
            return False
    return True